unit CompileU;

(*
//  C-to-Quadcode Compiler
//  ----------------------
//  Practical Work:
//      Compilers Summer Semester 2008
//      (Lecture by Mr. Schwinn)
//
//  Course:
//      TAI-06-ABC
//
//  Developers:
//     Kevin Niehage   (Matrikelnummer: 280888)
//     Michaela Hansel (Matrikelnummer: 221664)
*)

// IMPLEMENTATION STATUS: DONE
{.$DEFINE WRITE_COMMENTS}

interface

uses
  Classes,
  SysUtils,
  ConstsU,
  ErrorsU,
  ExpU,
  FileU,
  FuncsU,
  QuotesU,
  StmtsU,
  SwitchesU,
  SymbolsU,
  TypesU,
  VarsU;

procedure Compile(const AInputFileName : String; const AOutputFileName : String);

implementation

procedure Compile(const AInputFileName : String; const AOutputFileName : String);
  function IVTtoStr(const AIVT : TInterValueType) : String;
  begin
    Result := '';

    case AIVT of
      ivtNone  : Result := 'void';
      ivtInt   : Result := 'int';
      ivtFloat : Result := 'float';
    end;
  end;

  function GetResultType(const AInputA : TInterValueType; const AInputB : TInterValueType) : TInterValueType;
  begin
//    Result := ivtNone;

    if (AInputA = AInputB) then
      Result := AInputA
    else
    begin
      if ((AInputA = ivtInt) and (AInputB = ivtFloat)) or
         ((AInputA = ivtFloat) and (AInputB = ivtInt)) then
        Result := ivtFloat
      else
        raise TCompilerException.Create(CErrorTypeMismatch + IVTtoStr(AInputA) + ' <> ' + IVTtoStr(AInputB));
    end;
  end;

  function STTtoIVT(const ASTT : TSymbolTableType) : TInterValueType;
  begin
    Result := ivtNone;

    case ASTT of
      sttInt  : Result := ivtInt;
      sttReal : Result := ivtFloat;
    end;
  end;

  function GenerateExpression(const AExpInfo : TExpInfo; const ALineIndex : LongInt;
                              var AQuadCode : String; var ALineCount : LongInt;
                              const ASymbolTable : TSymbolTable; const AParentNr : LongInt;
                              var AFloatHelper : LongInt; var AIntHelper : LongInt) : TInterValue;
  var
    LLine : LongInt;

    function GetType(const AVariableName : String; const AFunctionName : String;
                     const ASymbolTable : TSymbolTable; const AParentNr : LongInt) : TInterValueType;
    var
      LSymbolIndex : LongInt;
    begin
      Result := ivtNone;

      if (ASymbolTable <> nil) and (AParentNr >= 0) then
      begin
        if (Length(AFunctionName) = 0) then
        begin
          LSymbolIndex := ASymbolTable.EntryExists(AVariableName, AParentNr);
          if (LSymbolIndex > 0) then
            Result := STTtoIVT(ASymbolTable.GetEntry(LSymbolIndex).Typ)
          else
          begin
            LSymbolIndex := ASymbolTable.EntryExists(AVariableName, 0);
            if (LSymbolIndex > 0) then
              Result := STTtoIVT(ASymbolTable.GetEntry(LSymbolIndex).Typ)
          end;
        end
        else
        begin
          if (Length(AVariableName) = 0) then
          begin
            LSymbolIndex := ASymbolTable.EntryExists(AFunctionName, 0);
            if (LSymbolIndex > 0) then
              Result := STTtoIVT(ASymbolTable.GetEntry(LSymbolIndex).IntTyp);
          end
          else
          begin
            LSymbolIndex := ASymbolTable.EntryExists(AFunctionName, 0);
            if (LSymbolIndex > 0) then
            begin
              LSymbolIndex := ASymbolTable.EntryExists(AVariableName, LSymbolIndex);
              if (LSymbolIndex > 0) then
              begin
                if (ASymbolTable.GetEntry(LSymbolIndex).Parameter > 0) then
                  Result := STTtoIVT(ASymbolTable.GetEntry(LSymbolIndex).Typ);
              end;
            end;
          end;
        end;
      end;
    end;

    function GetTypeHelper(const AType : TInterValueType) : TInterValue;
    begin
      Result.iConstruct := ivcVariable;
      Result.iType      := AType;
      Result.iValue     := '';

      case AType of
        ivtInt   : Result.iValue := CIntHelper + IntToStr(AIntHelper);
        ivtFloat : Result.iValue := CRealHelper + IntToStr(AFloatHelper);
      end;
    end;

    function GetNewTypeHelper(const AType : TInterValueType) : TInterValue;
    begin
      Result.iConstruct := ivcVariable;
      Result.iType      := AType;
      Result.iValue     := '';

      case AType of
        ivtInt :
        begin
          Inc(AIntHelper);
          Result := GetTypeHelper(AType);

          ASymbolTable.AddEntry(Result.iValue, sttInt, sttNone, 0, AParentNr, 0);
        end;

        ivtFloat :
        begin
          Inc(AFloatHelper);
          Result := GetTypeHelper(AType);

          ASymbolTable.AddEntry(Result.iValue, sttReal, sttNone, 0, AParentNr, 0);
        end;
      end;
    end;

    function GetValue(const AValue : TExpValue; const AFuncCalls : TInterValueArray; var AFuncIndex : LongInt) : TInterValue;
      function IndexOf(const AChar : Char; const ACharArray : array of Char) : LongInt;
      var
        LIndex : LongInt;
      begin
        Result := - 1;

        for LIndex := 0 to Pred(Length(ACharArray)) do
        begin
          if (AnsiLowerCase(ACharArray[LIndex]) = AnsiLowerCase(AChar)) then
          begin
            Result := LIndex;

            Break;
          end;
        end;
      end;

      function HexToDec(const AHex : String) : String;
      const
        CHexChars : array [0..15] of Char = ('0', '1', '2', '3', '4', '5', '6', '7',
                                             '8', '9', 'a', 'b', 'c', 'd', 'e', 'f');
      var
        LIndex  : LongInt;
        LResult : LongInt;
      begin
        LResult := 0;
        for LIndex := 1 to Length(AHex) do
          LResult := LResult * 16 + IndexOf(AHex[LIndex], CHexChars);

        Result := IntToStr(LResult);
      end;

      function OctToDec(const AOct : String) : String;
      const
        COctChars : array [0..7] of Char = ('0', '1', '2', '3', '4', '5', '6', '7');
      var
        LIndex  : LongInt;
        LResult : LongInt;
      begin
        LResult := 0;
        for LIndex := 1 to Length(AOct) do
          LResult := LResult * 8 + IndexOf(AOct[LIndex], COctChars);

        Result := IntToStr(LResult);
      end;
    begin
      Result.iConstruct := ivcNone;
      Result.iType      := ivtNone;
      Result.iValue     := '';

      case AValue.vType of
        evtHex :
        begin
          Result.iConstruct := ivcValue;
          Result.iType      := ivtInt;
          Result.iValue     := HexToDec(AValue.vValue);
        end;

        evtOct :
        begin
          Result.iConstruct := ivcValue;
          Result.iType      := ivtInt;
          Result.iValue     := OctToDec(AValue.vValue);
        end;

        evtFunctionCall :
        begin
          Result.iConstruct := ivcFuncCall;
          Result.iType      := GetType('', AValue.vValue, ASymbolTable, AParentNr);

          if ((AFuncIndex >= 0) and (AFuncIndex < Length(AFuncCalls))) then
          begin
            Result := AFuncCalls[AFuncIndex];
            Inc(AFuncIndex);
          end;
        end;

        evtInteger :
        begin
          Result.iConstruct := ivcValue;
          Result.iType      := ivtInt;
          Result.iValue     := AValue.vValue;
        end;

        evtFloat :
        begin
          Result.iConstruct := ivcValue;
          Result.iType      := ivtFloat;
          Result.iValue     := AValue.vValue;
        end;

        evtVariable :
        begin
          Result.iConstruct := ivcVariable;
          Result.iType      := GetType(AValue.vValue, '', ASymbolTable, AParentNr);
          Result.iValue     := AValue.vValue;
        end;
      end;
    end;

    procedure WriteCode(const ACode : String; const AWriteLineNr : Boolean);
    begin
      if AWriteLineNr then
      begin
        AQuadCode := AQuadCode + IntToStr(LLine) + #9;
        Inc(LLine);
      end;
      AQuadCode := AQuadCode + ACode;
    end;

    procedure EndLine(const AComment : String);
    begin
{$IF DEFINED(WRITE_COMMENTS)}
      AQuadCode := AQuadCode + ' // ' + AComment;
{$IFEND}
      AQuadCode := AQuadCode + sLineBreak;
    end;

    procedure WriteCodeLn(const ACode : String; const AComment : String);
    begin
      WriteCode(ACode, true);
      EndLine(AComment);
    end;

    procedure WriteAssign(const AResult : String; const AValue : String);
    begin
      WriteLn('        Generating Assignment');
      WriteCodeLn(AResult + ' := ' + AValue,
                  'assignment');
    end;

    procedure WriteParam(const AValue : String);
    begin
      WriteLn('        Generating Parameter Push');
      WriteCodeLn('PARAM ' + AValue,
                  'function call');
    end;

    procedure WriteFuncCall(const AResult : String; const AFuncName : String; const AParamCount : LongInt);
    begin
      WriteLn('        Generating Function Call');
      WriteAssign(AResult, 'CALL ' + AFuncName + ', ' + IntToStr(AParamCount));
    end;

    procedure WriteProcCall(const AFuncName : String; const AParamCount : LongInt);
    begin
      WriteLn('        Generating Procedure Call');
      WriteCodeLn('CALL ' + AFuncName + ', ' + IntToStr(AParamCount),
                  'procedure call');
    end;

    procedure WriteCalc(const AResult : String; const ALeft : String;
                        const ARight : String; const AOperation : String);
    begin
      WriteLn('        Generating Calculation');
      WriteAssign(AResult, ALeft + ' ' + AOperation + ' ' + ARight);
    end;

    procedure WriteNot(const AResult : String; const AValue : String);
    begin
      WriteLn('        Generating Not');
      WriteCodeLn('IF (' + AValue + ' = 0) GOTO ' + IntToStr(LLine + 3),
                  'not (switch)');
      WriteAssign(AResult, '0');
      WriteCodeLn('GOTO ' + IntToStr(LLine + 2),
                  'not (exit)');
      WriteAssign(AResult, '1');
    end;

    procedure WriteEquation(const AResult : String; const ALeft : String; const ARight : String;
                            const AOperation : String);
    begin
      WriteLn('        Generating Equation');
      WriteCodeLn('IF (' + ALeft + ' ' + AOperation + ' ' + ARight + ') GOTO ' + IntToStr(LLine + 3),
                  'equation (switch)');
      WriteAssign(AResult, '0');
      WriteCodeLn('GOTO ' + IntToStr(LLine + 2),
                  'equation (exit)');
      WriteAssign(AResult, '1');
    end;

    procedure WriteEqual(const AResult : String; const ALeft : String; const ARight : String);
    begin
      WriteLn('        Generating Equal');
      WriteEquation(AResult, ALeft, ARight, '=');
    end;

    procedure WriteNotEqual(const AResult : String; const ALeft : String; const ARight : String);
    begin
      WriteLn('        Generating Not Equal');
      WriteCodeLn('IF (' + ALeft + ' = ' + ARight + ') GOTO ' + IntToStr(LLine + 3),
                  'not equal (switch)');
      WriteAssign(AResult, '1');
      WriteCodeLn('GOTO ' + IntToStr(LLine + 2),
                  'not equal (exit)');
      WriteAssign(AResult, '0');
    end;

    procedure WriteAnd(const AResult : String; const ALeft : String; const ARight : String);
    begin
      WriteLn('        Generating And');
      WriteCodeLn('IF (' + ALeft + ' = 0) GOTO ' + IntToStr(LLine + 4),
                  'and (switch 1/2)');
      WriteCodeLn('IF (' + ARight + ' = 0) GOTO ' + IntToStr(LLine + 3),
                  'and (switch 2/2)');
      WriteAssign(AResult, '1');
      WriteCodeLn('GOTO ' + IntToStr(LLine + 2),
                  'not equal (exit)');
      WriteAssign(AResult, '0');
    end;

    procedure WriteOr(const AResult : String; const ALeft : String; const ARight : String);
    begin
      WriteLn('        Generating Or');
      WriteCodeLn('IF (' + ALeft + ' = 0) GOTO ' + IntToStr(LLine + 4),
                  'or (switch 1/4)');
      WriteCodeLn('IF (' + ARight + ' = 0) GOTO ' + IntToStr(LLine + 6),
                  'or (switch 2/4)');
      WriteAssign(AResult, '1');
      WriteCodeLn('GOTO ' + IntToStr(LLine + 8),
                  'or (exit 1/3)');
      WriteCodeLn('IF (' + ARight + ' = 0) GOTO ' + IntToStr(LLine + 6),
                  'or (switch 3/4)');
      WriteAssign(AResult, '1');
      WriteCodeLn('GOTO ' + IntToStr(LLine + 5),
                  'or (exit 2/3)');
      WriteCodeLn('IF (' + ALeft + ' = 0) GOTO ' + IntToStr(LLine + 3),
                  'or (switch 4/4)');
      WriteAssign(AResult, '1');
      WriteCodeLn('GOTO ' + IntToStr(LLine + 2),
                  'or (exit 3/3)');
      WriteAssign(AResult, '0');
    end;

    procedure WriteMod(const AResult : TInterValue; const ALeft : TInterValue; const ARight : TInterValue);
    var
      LLeft     : TInterValue;
      LRight    : TInterValue;
      LTemp     : String;
      LTempLine : String;
    begin
      WriteLn('        Generating Mod');
      LLeft  := GetNewTypeHelper(ALeft.iType);
      LRight := GetNewTypeHelper(ARight.iType);
      LTemp  := GetNewTypeHelper(ivtInt).iValue;

      if ((LLeft.iType = ivtInt) and (LRight.iType = ivtInt)) then
      begin
        WriteAssign(LLeft.iValue,  ALeft.iValue);
        WriteAssign(LRight.iValue, ARight.iValue);
        WriteCodeLn('IF (' + LRight.iValue + ' >= 0) GOTO ' + IntToStr(LLine + 2),
                    'mod (switch 1/3)');
        WriteCalc(LRight.iValue, '0', LRight.iValue, '-');
        WriteAssign(LTemp, '1');
        WriteCodeLn('IF (' + LLeft.iValue + ' >= 0) GOTO ' + IntToStr(LLine + 2),
                    'mod (switch 2/3)');
        WriteAssign(LTemp, '-1');
        WriteCalc(LLeft.iValue, LLeft.iValue, LTemp, '*');
        LTempLine := IntToStr(LLine);
        WriteCodeLn('IF (' + LLeft.iValue + ' < ' + LRight.iValue + ') GOTO ' + IntToStr(LLine + 3),
                    'mod (switch 3/3)');
        WriteCalc(LLeft.iValue, LLeft.iValue, LRight.iValue, '-');
        WriteCodeLn('GOTO ' + LTempLine,
                    'mod (loop)');
        WriteCalc(LLeft.iValue, LLeft.iValue, LTemp, '*');
        WriteAssign(AResult.iValue, LLeft.iValue);
      end
      else
        raise TCompilerException.Create(CErrorModuloOnNoInteger + '"' + ALeft.iValue + '" (' + IVTtoStr(ALeft.iType) + ') mod "' + ARight.iValue + '" (' + IVTtoStr(ARight.iType) + ')');
    end;

    procedure WriteShift(const AResult : TInterValue; const ALeft : TInterValue; const ARight : TInterValue;
                         const AShiftLeft : Boolean);
    var
      LLeft     : TInterValue;
      LOperator : String;
      LRight    : TInterValue;
      LTempLine : String;
    begin
      WriteLn('        Generating Shift');
      if AShiftLeft then
        LOperator := OperationToString(eoMulVal)
      else
        LOperator := OperationToString(eoDivVal);

      LLeft  := GetNewTypeHelper(ALeft.iType);
      LRight := GetNewTypeHelper(ARight.iType);

      if ((LLeft.iType = ivtInt) and (LRight.iType = ivtInt)) then
      begin
        WriteAssign(LLeft.iValue,  ALeft.iValue);
        WriteAssign(LRight.iValue, ARight.iValue);
        LTempLine := IntToStr(LLine);
        WriteCodeLn('IF (' + LRight.iValue + ' <= 0) GOTO ' + IntToStr(LLine + 4),
                    'shift (switch)');
        WriteCalc(LLeft.iValue, LLeft.iValue, '2', LOperator);
        WriteCalc(LRight.iValue, LRight.iValue, '1', '-');
        WriteCodeLn('GOTO ' + LTempLine,
                    'shift (loop)');
        WriteAssign(AResult.iValue, LLeft.iValue);
      end
      else
        raise TCompilerException.Create(CErrorShiftOnNoInteger + '"' + ALeft.iValue + '" (' + IVTtoStr(ALeft.iType) + ') mod "' + ARight.iValue + '" (' + IVTtoStr(ARight.iType) + ')');
    end;

    function GenerateExp(const ATree : PExpEntry; const AIsOnLeft : Boolean;
                         const AAbove : TInterValue) : TInterValue;
    var
      LBefore     : TInterValue;
      LCode       : String;
      LFuncCalls  : TInterValueArray;
      LFuncIndex  : LongInt;
      LIndex      : LongInt;
      LIndexB     : LongInt;
      LLines      : LongInt;
      LParams     : TInterValueArray;
      LParamTypes : TSymbolTableTypeArray;
      LTemp       : TInterValue;
      LValueA     : TInterValue;
      LValueB     : TInterValue;
    begin
      LBefore.iConstruct := ivcNone;
      LBefore.iType      := ivtNone;
      LBefore.iValue     := '';

      if (ATree <> nil) then
      begin
        if (ATree^.eRight <> nil) then
          LBefore := GenerateExp(ATree^.eRight, false, LBefore);

        LFuncIndex := -1;
        SetLength(LParamTypes, 0);
        for LIndex := 0 to Pred(Length(ATree^.eValues)) do
        begin
          LValueA := GetValue(ATree^.eValues[LIndex], nil, LFuncIndex);

          if (LValueA.iConstruct = ivcFuncCall) then
          begin
            SetLength(LFuncCalls, Succ(Length(LFuncCalls)));

            SetLength(LParams, Length(ATree^.eValues[LIndex].vParams));
            LParamTypes := ASymbolTable.ParamTypeList(ATree^.eValues[LIndex].vValue, 0);

            for LIndexB := 0 to Pred(Length(ATree^.eValues[LIndex].vParams)) do
            begin
              LParams[LIndexB] := GenerateExpression(ATree^.eValues[LIndex].vParams[LIndexB],
                                                     LLine, LCode, LLines, ASymbolTable, AParentNr,
                                                     AFloatHelper, AIntHelper);

              if (STTtoIVT(LParamTypes[LIndexB]) = LParams[LIndexB].iType) then
              begin
                AQuadCode := AQuadCode + LCode;
                LLine     := LLine + LLines;
              end
              else
              begin
                if (LParams[LIndexB].iType = ivtInt) and
                   (STTtoIVT(LParamTypes[LIndexB]) = ivtFloat) then
                begin
                  AQuadCode := AQuadCode + LCode;
                  LLine     := LLine + LLines;

                  WriteAssign(GetNewTypeHelper(ivtFloat).iValue, LParams[LIndexB].iValue);
                  LParams[LIndexB].iValue := GetTypeHelper(ivtFloat).iValue;
                end
                else
                  raise TCompilerException.Create(CErrorParamTypeMismatch + IVTtoStr(LParams[LIndexB].iType) + '<>' + IVTtoStr(STTtoIVT(LParamTypes[LIndexB])));
              end;
            end;

            for LIndexB := 0 to Pred(Length(LParams)) do
              WriteParam(LParams[LIndexB].iValue);

            LFuncCalls[High(LFuncCalls)] := GetNewTypeHelper(GetType('', ATree^.eValues[LIndex].vValue, ASymbolTable, AParentNr));
            if (LFuncCalls[High(LFuncCalls)].iType = ivtNone) then
              WriteProcCall(ATree^.eValues[LIndex].vValue, Length(LParams))
            else
              WriteFuncCall(LFuncCalls[High(LFuncCalls)].iValue, ATree^.eValues[LIndex].vValue, Length(LParams));
          end;
        end;

        LFuncIndex := 0;
        for LIndex := 0 to Pred(Length(ATree^.eValues)) do
        begin
          if (ATree^.eValues[LIndex].vOption <> evoNone) then
          begin
            LValueA := GetValue(ATree^.eValues[LIndex], LFuncCalls, LFuncIndex);

            if (LValueA.iConstruct = ivcVariable) then
            begin
              case ATree^.eValues[LIndex].vOption of
                evoPreDec :
                begin
                  if (LValueA.iType = ivtInt) then
                    WriteCalc(LValueA.iValue, LValueA.iValue, '1', '-')
                  else
                    raise TCompilerException.Create(CErrorIncrementOnNoInteger + '"' + LValueA.iValue + '" (' + IVTtoStr(LValueA.iType) + ')');
                end;

                evoPreInc :
                begin
                  if (LValueA.iType = ivtInt) then
                    WriteCalc(LValueA.iValue, LValueA.iValue, '1', '+')
                  else
                    raise TCompilerException.Create(CErrorIncrementOnNoInteger + '"' + LValueA.iValue + '" (' + IVTtoStr(LValueA.iType) + ')');
                end;
              end;
            end
            else
              raise TCompilerException.Create(CErrorIncrementOnNoVariable + LValueA.iValue);
          end;
        end;

        case ATree^.eOperation of
          eoModVal :
          begin
            for LIndex := 0 to Pred(Length(ATree^.eValues)) do
            begin
              LValueA := GetValue(ATree^.eValues[LIndex], LFuncCalls, LFuncIndex);

              if (LIndex = 0) then
              begin
                if AIsOnLeft then
                begin
                  LTemp := GetNewTypeHelper(GetResultType(AAbove.iType, LValueA.iType));
                  WriteMod(LTemp, AAbove, LValueA);
                  LValueA := LTemp;
                end;

                if (ATree^.eRight <> nil) then
                begin
                  LTemp := GetNewTypeHelper(GetResultType(LValueA.iType, LBefore.iType));
                  WriteMod(LTemp, LBefore, LValueA);
                  LValueA := LTemp;
                end;
              end
              else
              begin
                LTemp := GetNewTypeHelper(GetResultType(LValueA.iType, LBefore.iType));
                WriteMod(LTemp, LBefore, LValueA);
                LValueA := LTemp;
              end;

              LBefore := LValueA;
            end;

            if (Length(ATree^.eValues) <= 0) then
            begin
              if ((ATree^.eRight <> nil) and AIsOnLeft) then
              begin
                LTemp := GetNewTypeHelper(GetResultType(AAbove.iType, LBefore.iType));
                WriteMod(LTemp, AAbove, LBefore);
                LBefore := LTemp;
              end
              else
                raise TCompilerException.Create(CErrorInvalidOperationEvaluation);
            end;
          end;

          eoShlVal,
          eoShrVal :
          begin
            for LIndex := 0 to Pred(Length(ATree^.eValues)) do
            begin
              LValueA := GetValue(ATree^.eValues[LIndex], LFuncCalls, LFuncIndex);

              if (LIndex = 0) then
              begin
                if AIsOnLeft then
                begin
                  LTemp := GetNewTypeHelper(GetResultType(AAbove.iType, LValueA.iType));
                  WriteShift(LTemp, AAbove, LValueA, (ATree^.eOperation = eoShlVal));
                  LValueA := LTemp;
                end;

                if (ATree^.eRight <> nil) then
                begin
                  LTemp := GetNewTypeHelper(GetResultType(LValueA.iType, LBefore.iType));
                  WriteShift(LTemp, LBefore, LValueA, (ATree^.eOperation = eoShlVal));
                  LValueA := LTemp;
                end;
              end
              else
              begin
                LTemp := GetNewTypeHelper(GetResultType(LValueA.iType, LBefore.iType));
                WriteShift(LTemp, LBefore, LValueA, (ATree^.eOperation = eoShlVal));
                LValueA := LTemp;
              end;

              LBefore := LValueA;
            end;

            if (Length(ATree^.eValues) <= 0) then
            begin
              if ((ATree^.eRight <> nil) and AIsOnLeft) then
              begin
                LTemp := GetNewTypeHelper(GetResultType(AAbove.iType, LBefore.iType));
                WriteShift(LTemp, AAbove, LBefore, (ATree^.eOperation = eoShlVal));
                LBefore := LTemp;
              end
              else
                raise TCompilerException.Create(CErrorInvalidOperationEvaluation);
            end;
          end;

          eoDivVal,
          eoMulVal,
          eoMinusVal,
          eoPlusVal :
          begin
            for LIndex := 0 to Pred(Length(ATree^.eValues)) do
            begin
              LValueA := GetValue(ATree^.eValues[LIndex], LFuncCalls, LFuncIndex);

              if (LIndex = 0) then
              begin
                if AIsOnLeft then
                begin
                  WriteCalc(GetNewTypeHelper(GetResultType(AAbove.iType, LValueA.iType)).iValue,
                            AAbove.iValue, LValueA.iValue, OperationToString(ATree^.eOperation));
                  LValueA := GetTypeHelper(GetResultType(AAbove.iType, LValueA.iType));
                end;

                if (ATree^.eRight <> nil) then
                begin
                  WriteCalc(GetNewTypeHelper(GetResultType(LValueA.iType, LBefore.iType)).iValue,
                            LBefore.iValue, LValueA.iValue, OperationToString(ATree^.eOperation));
                  LValueA := GetTypeHelper(GetResultType(LValueA.iType, LBefore.iType));
                end;
              end
              else
              begin
                WriteCalc(GetNewTypeHelper(GetResultType(LValueA.iType, LBefore.iType)).iValue,
                          LBefore.iValue, LValueA.iValue, OperationToString(ATree^.eOperation));
                LValueA := GetTypeHelper(GetResultType(LValueA.iType, LBefore.iType));
              end;

              LBefore := LValueA;
            end;

            if (Length(ATree^.eValues) <= 0) then
            begin
              if ((ATree^.eRight <> nil) and AIsOnLeft) then
              begin
                WriteCalc(GetNewTypeHelper(GetResultType(AAbove.iType, LBefore.iType)).iValue,
                          AAbove.iValue, LBefore.iValue, OperationToString(ATree^.eOperation));
                LBefore := GetTypeHelper(GetResultType(AAbove.iType, LBefore.iType));
              end
              else
                raise TCompilerException.Create(CErrorInvalidOperationEvaluation);
            end;
          end;

          eoNotLog :
          begin
            if (ATree^.eRight <> nil) then
            begin
              LValueA := LBefore;

              WriteNot(GetNewTypeHelper(LBefore.iType).iValue, LBefore.iValue);
              LBefore := GetTypeHelper(LBefore.iType);
            end
            else
            begin
              if (Length(ATree^.eValues) = 1) then
              begin
                LValueA := GetValue(ATree^.eValues[0], LFuncCalls, LFuncIndex);

                WriteNot(GetNewTypeHelper(LValueA.iType).iValue, LValueA.iValue);
                LBefore := GetTypeHelper(LValueA.iType);
              end
              else
                raise TCompilerException.Create(CErrorUnaryOperationWithWrongNumberOfOperands + IntToStr(Length(ATree^.eValues)));
            end;
          end;

          eoEqualLog :
          begin
            for LIndex := 0 to Pred(Length(ATree^.eValues)) do
            begin
              LValueA := GetValue(ATree^.eValues[LIndex], LFuncCalls, LFuncIndex);

              if (LIndex = 0) then
              begin
                if AIsOnLeft then
                begin
                  WriteEqual(GetNewTypeHelper(GetResultType(AAbove.iType, LValueA.iType)).iValue,
                             AAbove.iValue, LValueA.iValue);
                  LValueA := GetTypeHelper(GetResultType(AAbove.iType, LValueA.iType));
                end;

                if (ATree^.eRight <> nil) then
                begin
                  WriteEqual(GetNewTypeHelper(GetResultType(LValueA.iType, LBefore.iType)).iValue,
                             LBefore.iValue, LValueA.iValue);
                  LValueA := GetTypeHelper(GetResultType(LValueA.iType, LBefore.iType));
                end;
              end
              else
              begin
                WriteEqual(GetNewTypeHelper(GetResultType(LValueA.iType, LBefore.iType)).iValue,
                          LBefore.iValue, LValueA.iValue);
                LValueA := GetTypeHelper(GetResultType(LValueA.iType, LBefore.iType));
              end;

              LBefore := LValueA;
            end;

            if (Length(ATree^.eValues) <= 0) then
            begin
              if ((ATree^.eRight <> nil) and AIsOnLeft) then
              begin
                WriteEqual(GetNewTypeHelper(GetResultType(AAbove.iType, LBefore.iType)).iValue,
                          AAbove.iValue, LBefore.iValue);
                LBefore := GetTypeHelper(GetResultType(AAbove.iType, LBefore.iType));
              end
              else
                raise TCompilerException.Create(CErrorInvalidOperationEvaluation);
            end;
          end;

          eoNotEqualLog :
          begin
            for LIndex := 0 to Pred(Length(ATree^.eValues)) do
            begin
              LValueA := GetValue(ATree^.eValues[LIndex], LFuncCalls, LFuncIndex);

              if (LIndex = 0) then
              begin
                if AIsOnLeft then
                begin
                  WriteNotEqual(GetNewTypeHelper(GetResultType(AAbove.iType, LValueA.iType)).iValue,
                                AAbove.iValue, LValueA.iValue);
                  LValueA := GetTypeHelper(GetResultType(AAbove.iType, LValueA.iType));
                end;

                if (ATree^.eRight <> nil) then
                begin
                  WriteNotEqual(GetNewTypeHelper(GetResultType(LValueA.iType, LBefore.iType)).iValue,
                                LBefore.iValue, LValueA.iValue);
                  LValueA := GetTypeHelper(GetResultType(LValueA.iType, LBefore.iType));
                end;
              end
              else
              begin
                WriteNotEqual(GetNewTypeHelper(GetResultType(LValueA.iType, LBefore.iType)).iValue,
                              LBefore.iValue, LValueA.iValue);
                LValueA := GetTypeHelper(GetResultType(LValueA.iType, LBefore.iType));
              end;

              LBefore := LValueA;
            end;

            if (Length(ATree^.eValues) <= 0) then
            begin
              if ((ATree^.eRight <> nil) and AIsOnLeft) then
              begin
                WriteNotEqual(GetNewTypeHelper(GetResultType(AAbove.iType, LBefore.iType)).iValue,
                              AAbove.iValue, LBefore.iValue);
                LBefore := GetTypeHelper(GetResultType(AAbove.iType, LBefore.iType));
              end
              else
                raise TCompilerException.Create(CErrorInvalidOperationEvaluation);
            end;
          end;

          eoGreaterEqualLog,
          eoGreaterLog,
          eoLessEqualLog,
          eoLessLog :
          begin
            for LIndex := 0 to Pred(Length(ATree^.eValues)) do
            begin
              LValueA := GetValue(ATree^.eValues[LIndex], LFuncCalls, LFuncIndex);

              if (LIndex = 0) then
              begin
                if AIsOnLeft then
                begin
                  WriteEquation(GetNewTypeHelper(GetResultType(AAbove.iType, LValueA.iType)).iValue,
                                AAbove.iValue, LValueA.iValue, OperationToString(ATree^.eOperation));
                  LValueA := GetTypeHelper(GetResultType(AAbove.iType, LValueA.iType));
                end;

                if (ATree^.eRight <> nil) then
                begin
                  WriteEquation(GetNewTypeHelper(GetResultType(LValueA.iType, LBefore.iType)).iValue,
                                LBefore.iValue, LValueA.iValue, OperationToString(ATree^.eOperation));
                  LValueA := GetTypeHelper(GetResultType(LValueA.iType, LBefore.iType));
                end;
              end
              else
              begin
                WriteEquation(GetNewTypeHelper(GetResultType(LValueA.iType, LBefore.iType)).iValue,
                              LBefore.iValue, LValueA.iValue, OperationToString(ATree^.eOperation));
                LValueA := GetTypeHelper(GetResultType(LValueA.iType, LBefore.iType));
              end;

              LBefore := LValueA;
            end;

            if (Length(ATree^.eValues) <= 0) then
            begin
              if ((ATree^.eRight <> nil) and AIsOnLeft) then
              begin
                WriteEquation(GetNewTypeHelper(GetResultType(AAbove.iType, LBefore.iType)).iValue,
                              AAbove.iValue, LBefore.iValue, OperationToString(ATree^.eOperation));
                LBefore := GetTypeHelper(GetResultType(AAbove.iType, LBefore.iType));
              end
              else
                raise TCompilerException.Create(CErrorInvalidOperationEvaluation);
            end;
          end;

          eoAndLog :
          begin
            for LIndex := 0 to Pred(Length(ATree^.eValues)) do
            begin
              LValueA := GetValue(ATree^.eValues[LIndex], LFuncCalls, LFuncIndex);

              if (LIndex = 0) then
              begin
                if AIsOnLeft then
                begin
                  WriteAnd(GetNewTypeHelper(GetResultType(AAbove.iType, LValueA.iType)).iValue,
                           AAbove.iValue, LValueA.iValue);
                  LValueA := GetTypeHelper(GetResultType(AAbove.iType, LValueA.iType));
                end;

                if (ATree^.eRight <> nil) then
                begin
                  WriteAnd(GetNewTypeHelper(GetResultType(LValueA.iType, LBefore.iType)).iValue,
                           LBefore.iValue, LValueA.iValue);
                  LValueA := GetTypeHelper(GetResultType(LValueA.iType, LBefore.iType));
                end;
              end
              else
              begin
                WriteAnd(GetNewTypeHelper(GetResultType(LValueA.iType, LBefore.iType)).iValue,
                         LBefore.iValue, LValueA.iValue);
                LValueA := GetTypeHelper(GetResultType(LValueA.iType, LBefore.iType));
              end;

              LBefore := LValueA;
            end;

            if (Length(ATree^.eValues) <= 0) then
            begin
              if ((ATree^.eRight <> nil) and AIsOnLeft) then
              begin
                WriteAnd(GetNewTypeHelper(GetResultType(AAbove.iType, LBefore.iType)).iValue,
                         AAbove.iValue, LBefore.iValue);
                LBefore := GetTypeHelper(GetResultType(AAbove.iType, LBefore.iType));
              end
              else
                raise TCompilerException.Create(CErrorInvalidOperationEvaluation);
            end;
          end;

          eoOrLog :
          begin
            for LIndex := 0 to Pred(Length(ATree^.eValues)) do
            begin
              LValueA := GetValue(ATree^.eValues[LIndex], LFuncCalls, LFuncIndex);

              if (LIndex = 0) then
              begin
                if AIsOnLeft then
                begin
                  WriteOr(GetNewTypeHelper(GetResultType(AAbove.iType, LValueA.iType)).iValue,
                          AAbove.iValue, LValueA.iValue);
                  LValueA := GetTypeHelper(GetResultType(AAbove.iType, LValueA.iType));
                end;

                if (ATree^.eRight <> nil) then
                begin
                  WriteOr(GetNewTypeHelper(GetResultType(LValueA.iType, LBefore.iType)).iValue,
                          LBefore.iValue, LValueA.iValue);
                  LValueA := GetTypeHelper(GetResultType(LValueA.iType, LBefore.iType));
                end;
              end
              else
              begin
                WriteOr(GetNewTypeHelper(GetResultType(LValueA.iType, LBefore.iType)).iValue,
                        LBefore.iValue, LValueA.iValue);
                LValueA := GetTypeHelper(GetResultType(LValueA.iType, LBefore.iType));
              end;

              LBefore := LValueA;
            end;

            if (Length(ATree^.eValues) <= 0) then
            begin
              if ((ATree^.eRight <> nil) and AIsOnLeft) then
              begin
                WriteOr(GetNewTypeHelper(GetResultType(AAbove.iType, LBefore.iType)).iValue,
                        AAbove.iValue, LBefore.iValue);
                LBefore := GetTypeHelper(GetResultType(AAbove.iType, LBefore.iType));
              end
              else
                raise TCompilerException.Create(CErrorInvalidOperationEvaluation);
            end;
          end;

          eoAssign :
          begin
            if not(AIsOnLeft) then
            begin
              for LIndex := Pred(Length(ATree^.eValues)) downto 0 do
              begin
                LValueA := GetValue(ATree^.eValues[LIndex], LFuncCalls, LFuncIndex);

                if (LIndex = High(ATree^.eValues)) then
                begin
                  if (ATree^.eRight <> nil) then
                  begin
                    if (LValueA.iConstruct = ivcVariable) then
                    begin
                      if (GetResultType(LValueA.iType, LBefore.iType) = LValueA.iType) then
                        WriteAssign(LValueA.iValue, LBefore.iValue)
                      else
                        raise TCompilerException.Create(CErrorTypeMismatch + IVTtoStr(LBefore.iType) + '<>' + IVTtoStr(LValueA.iType));
                    end
                    else
                      raise TCompilerException.Create(CErrorAssignmentOnNoVariable + LValueA.iValue);
                  end;
                end
                else
                begin
                  if (LValueA.iConstruct = ivcVariable) then
                  begin
//                    LValueB := GetValue(ATree^.eValues[Succ(LIndex)], LFuncCalls, LFuncIndex);

                    if (GetResultType(LValueA.iType, LBefore.iType) = LValueA.iType) then
                      WriteAssign(LValueA.iValue, LBefore.iValue)
                    else
                      raise TCompilerException.Create(CErrorTypeMismatch + IVTtoStr(LValueA.iType) + '<>' + IVTtoStr(LBefore.iType));
                  end
                  else
                    raise TCompilerException.Create(CErrorAssignmentOnNoVariable + LValueA.iValue);
                end;

                LBefore := LValueA;
              end;
            end
            else
              raise TCompilerException.Create(CErrorAssignmentWithPrecedingOperation);
          end;
        else
          raise TCompilerException.Create(CErrorUnsupportedOperation + OperationToString(ATree^.eOperation));
        end;

        for LIndex := 0 to Pred(Length(ATree^.eValues)) do
        begin
          if (ATree^.eValues[LIndex].vOption <> evoNone) then
          begin
            LValueA := GetValue(ATree^.eValues[LIndex], LFuncCalls, LFuncIndex);

            if (LValueA.iConstruct = ivcVariable) then
            begin
              case ATree^.eValues[LIndex].vOption of
                evoPostDec :
                begin
                  if (LValueA.iType = ivtInt) then
                    WriteCalc(LValueA.iValue, LValueA.iValue, '1', '-')
                  else
                    raise TCompilerException.Create(CErrorIncrementOnNoInteger + '"' + LValueA.iValue + '" (' + IVTtoStr(LValueA.iType) + ')');
                end;

                evoPostInc :
                begin
                  if (LValueA.iType = ivtInt) then
                    WriteCalc(LValueA.iValue, LValueA.iValue, '1', '+')
                  else
                    raise TCompilerException.Create(CErrorIncrementOnNoInteger + '"' + LValueA.iValue + '" (' + IVTtoStr(LValueA.iType) + ')');
                end;
              end;
            end
            else
              raise TCompilerException.Create(CErrorIncrementOnNoVariable + LValueA.iValue);
          end;
        end;

        if (ATree^.eLeft <> nil) then
          LBefore := GenerateExp(ATree^.eLeft, true, LBefore);
      end;

      Result := LBefore;
    end;
  begin
    Result.iConstruct := ivcNone;
    Result.iType      := ivtNone;
    Result.iValue     := '';

    AQuadCode := '';

    LLine := ALineIndex;
    if (AExpInfo.eEntry <> nil) then
      Result := GenerateExp(AExpInfo.eEntry, false, Result);
    ALineCount := LLine - ALineIndex;
  end;

  procedure CompileStatement(const AStatement : PStatementInfo; var ALineIndex : LongInt;
                             var AOutputSource : String; const ASymbolTable : TSymbolTable;
                             const ASymbolNr : LongInt; const AFuncHeaderInfos : TFuncHeaderInfoArray;
                             const AFuncHeaderInfoIndex : LongInt; var AFloatHelper : LongInt;
                             var AIntHelper : LongInt);
  var
    LIndex           : LongInt;
    LInsertLn        : LongInt;
    LInsertPos       : LongInt;
    LLineCount       : LongInt;
    LNewOutputSource : String;
    LResult          : TInterValue;
    LStart           : LongInt;
    LSymbols         : TSymbolArray;
    LTempExpression  : TExpInfo;
  begin
    SetLength(LSymbols, 0);

    if (AStatement <> nil) then
    begin
      case AStatement^.sType of
        sitBlock :
        begin
          WriteLn('    Compiling Block Statement');
          for LIndex := 0 to Pred(Length(AStatement^.sStatements)) do
            CompileStatement(AStatement^.sStatements[LIndex], ALineIndex,
                             AOutputSource, ASymbolTable, ASymbolNr,
                             AFuncHeaderInfos, AFuncHeaderInfoIndex,
                             AFloatHelper, AIntHelper);
        end;

        sitVariable :
        begin
          for LIndex := 0 to Pred(Length(AStatement^.sVariables)) do
          begin
            WriteLn('    Compiling Variable Statement: ' + AStatement^.sVariables[LIndex].vName);
            if (AStatement^.sVariables[LIndex].vType <> sttNone) then
            begin
              WriteLn('      Writing Variable to Symbol Table');
              ASymbolTable.AddEntry(AStatement^.sVariables[LIndex].vName,
                                    AStatement^.sVariables[LIndex].vType,
                                    sttNone, 0, ASymbolNr, 0);

              WriteLn('      Generating Source');
              LResult := GenerateExpression(AStatement^.sVariables[LIndex].vValue,
                                            ALineIndex, LNewOutputSource,
                                            LLineCount, ASymbolTable, ASymbolNr,
                                            AFloatHelper, AIntHelper);
              ALineIndex    := ALineIndex + LLineCount;
              AOutputSource := AOutputSource + LNewOutputSource;

              WriteLn('      Writing Assignment');
              LTempExpression := ParseExpression(AStatement^.sVariables[LIndex].vName + ' = ' + LResult.iValue);
              WriteLn('      Generating Assignment Source');
              GenerateExpression(LTempExpression, ALineIndex, LNewOutputSource,
                                 LLineCount, ASymbolTable, ASymbolNr, AFloatHelper, AIntHelper);
              ALineIndex    := ALineIndex + LLineCount;
              AOutputSource := AOutputSource + LNewOutputSource;
            end
            else
              raise TCompilerException.Create(CErrorVoidVariable + AStatement^.sVariables[LIndex].vName);
          end;
        end;

        sitExpression :
        begin
          WriteLn('    Compiling Expression');
          LSymbols := GetSymbols(AStatement^.sExpression);
          WriteLn('      Checking if All Symbols Existing');
          for LIndex := 0 to Pred(Length(LSymbols)) do
          begin
            if (LSymbols[LIndex].sType = stFunction) then
            begin
              if (ASymbolTable.EntryExists(LSymbols[LIndex].sName, 0) < 0) then
                raise TCompilerException.Create(CErrorUnknownSymbol + AFuncHeaderInfos[AFuncHeaderInfoIndex].fName);
            end
            else
            begin
              if (ASymbolTable.EntryExists(LSymbols[LIndex].sName, ASymbolNr) < 0) then
              begin
                if (ASymbolTable.EntryExists(LSymbols[LIndex].sName, 0) < 0) then
                  raise TCompilerException.Create(CErrorUnknownSymbol + AFuncHeaderInfos[AFuncHeaderInfoIndex].fName);
              end;
            end;
          end;

          WriteLn('      Generating Source');
          GenerateExpression(AStatement^.sExpression, ALineIndex, LNewOutputSource,
                             LLineCount, ASymbolTable, ASymbolNr, AFloatHelper, AIntHelper);
          ALineIndex    := ALineIndex + LLineCount;
          AOutputSource := AOutputSource + LNewOutputSource;
        end;

        sitWhile :
        begin
          WriteLn('    Compiling While Statement');
          if (Length(AStatement^.sStatements) = 1) then
          begin
            LStart := ALineIndex;

            WriteLn('      Generating Source');
            LResult := GenerateExpression(AStatement^.sExpression,
                                          ALineIndex, LNewOutputSource,
                                          LLineCount, ASymbolTable, ASymbolNr,
                                          AFloatHelper, AIntHelper);
            ALineIndex    := ALineIndex + LLineCount;
            AOutputSource := AOutputSource + LNewOutputSource;

            LInsertLn := ALineIndex;
            Inc(ALineIndex);
            LInsertPos := Length(AOutputSource);

            WriteLn('      Compiling Statement');
            CompileStatement(AStatement^.sStatements[0], ALineIndex, AOutputSource,
                             ASymbolTable, ASymbolNr, AFuncHeaderInfos, AFuncHeaderInfoIndex,
                             AFloatHelper, AIntHelper);

            WriteLn('      Injecting Jump Address');
            AOutputSource := Copy(AOutputSource, 1, LInsertPos) +
                             IntToStr(LInsertLn) + #9 + 'IF (' + LResult.iValue + ' = 0) GOTO ' + IntToStr(Succ(ALineIndex)) + sLineBreak +
                             Copy(AOutputSource, Succ(LInsertPos), Length(AOutputSource) - LInsertPos) +
                             IntToStr(ALineIndex) + #9 + 'GOTO ' + IntToStr(LStart) + sLineBreak;
            Inc(ALineIndex);
          end
          else
            raise TCompilerException.Create(CErrorWhileWithoutExpression);
        end;

        sitDo :
        begin
          WriteLn('    Compiling Do Statement');
          if (Length(AStatement^.sStatements) = 1) then
          begin
            LStart := ALineIndex;

            WriteLn('      Compiling Statement');
            CompileStatement(AStatement^.sStatements[0], ALineIndex, AOutputSource,
                             ASymbolTable, ASymbolNr, AFuncHeaderInfos, AFuncHeaderInfoIndex,
                             AFloatHelper, AIntHelper);

            WriteLn('      Generating Source');
            LResult := GenerateExpression(AStatement^.sExpression,
                                          ALineIndex, LNewOutputSource,
                                          LLineCount, ASymbolTable, ASymbolNr,
                                          AFloatHelper, AIntHelper);
            ALineIndex    := ALineIndex + LLineCount;
            AOutputSource := AOutputSource + LNewOutputSource;

            WriteLn('      Injecting Jump Address');
            AOutputSource := AOutputSource + IntToStr(ALineIndex) + #9 + 'IF (' + LResult.iValue + ' = 0) GOTO ' + IntToStr(ALineIndex + 2) + sLineBreak;
            Inc(ALineIndex);
            AOutputSource := AOutputSource + IntToStr(ALineIndex) + #9 + 'GOTO ' + IntToStr(LStart) + sLineBreak;
            Inc(ALineIndex);
          end
          else
            raise TCompilerException.Create(CErrorDoWithoutExpression);
        end;

        sitIf :
        begin
          WriteLn('    Compiling If Statement');
          if (Length(AStatement^.sStatements) = 1) then
          begin
            WriteLn('      Generating Source');
            LResult := GenerateExpression(AStatement^.sExpression,
                                          ALineIndex, LNewOutputSource,
                                          LLineCount, ASymbolTable, ASymbolNr,
                                          AFloatHelper, AIntHelper);
            ALineIndex    := ALineIndex + LLineCount;
            AOutputSource := AOutputSource + LNewOutputSource;

            LInsertLn := ALineIndex;
            Inc(ALineIndex);
            LInsertPos := Length(AOutputSource);

            WriteLn('      Compiling Statement');
            CompileStatement(AStatement^.sStatements[0], ALineIndex, AOutputSource,
                             ASymbolTable, ASymbolNr, AFuncHeaderInfos, AFuncHeaderInfoIndex,
                             AFloatHelper, AIntHelper);

            WriteLn('      Injecting Jump Address');
            AOutputSource := Copy(AOutputSource, 1, LInsertPos) +
                             IntToStr(LInsertLn) + #9 + 'IF (' + LResult.iValue + ' = 0) GOTO ' + IntToStr(Succ(ALineIndex)) + sLineBreak +
                             Copy(AOutputSource, Succ(LInsertPos), Length(AOutputSource) - LInsertPos);

            LInsertLn := ALineIndex;
            Inc(ALineIndex);
            LInsertPos := Length(AOutputSource);

            WriteLn('      Compiling Else Statement');
            CompileStatement(AStatement^.sElseStatements, ALineIndex, AOutputSource,
                             ASymbolTable, ASymbolNr, AFuncHeaderInfos, AFuncHeaderInfoIndex,
                             AFloatHelper, AIntHelper);

            WriteLn('      Injecting Jump Address');
            AOutputSource := Copy(AOutputSource, 1, LInsertPos) +
                             IntToStr(LInsertLn) + #9 + 'GOTO ' + IntToStr(ALineIndex) + sLineBreak +
                             Copy(AOutputSource, Succ(LInsertPos), Length(AOutputSource) - LInsertPos);
          end
          else
            raise TCompilerException.Create(CErrorIfWithoutExpression);
        end;

        sitReturn :
        begin
          WriteLn('    Compiling Return Statement');
          LSymbols := GetSymbols(AStatement^.sExpression);
          for LIndex := 0 to Pred(Length(LSymbols)) do
          begin
            if (LSymbols[LIndex].sType = stFunction) then
            begin
              if (ASymbolTable.EntryExists(LSymbols[LIndex].sName, 0) < 0) then
                raise TCompilerException.Create(CErrorUnknownSymbol + AFuncHeaderInfos[AFuncHeaderInfoIndex].fName);
            end
            else
            begin
              if (ASymbolTable.EntryExists(LSymbols[LIndex].sName, ASymbolNr) < 0) then
              begin
                if (ASymbolTable.EntryExists(LSymbols[LIndex].sName, 0) < 0) then
                  raise TCompilerException.Create(CErrorUnknownSymbol + AFuncHeaderInfos[AFuncHeaderInfoIndex].fName);
              end;
            end;
          end;

          if (AFuncHeaderInfos[AFuncHeaderInfoIndex].fType = sttNone) then
          begin
            if (AStatement^.sExpression.eEntry = nil) then
            begin
              AOutputSource := AOutputSource + IntToStr(ALineIndex) + #9 + 'RETURN' + sLineBreak;
              Inc(ALineIndex);
            end
            else
              raise TCompilerException.Create(CErrorReturnExpressionForVoid);
          end
          else
          begin
            if (AStatement^.sExpression.eEntry <> nil) then
            begin
              WriteLn('      Generating Source');
              LResult := GenerateExpression(AStatement^.sExpression,
                                            ALineIndex, LNewOutputSource,
                                            LLineCount, ASymbolTable, ASymbolNr,
                                            AFloatHelper, AIntHelper);

              if (GetResultType(LResult.iType, STTtoIVT(AFuncHeaderInfos[AFuncHeaderInfoIndex].fType)) = STTtoIVT(AFuncHeaderInfos[AFuncHeaderInfoIndex].fType)) then
              begin
                ALineIndex    := ALineIndex + LLineCount;
                AOutputSource := AOutputSource + LNewOutputSource;

                AOutputSource := AOutputSource + IntToStr(ALineIndex) + #9 + 'RETURN ' + LResult.iValue + sLineBreak;
                Inc(ALineIndex);
              end
              else
                raise TCompilerException.Create(CErrorTypeMismatch + IVTtoStr(STTtoIVT(AFuncHeaderInfos[AFuncHeaderInfoIndex].fType)) + '<>' + IVTtoStr(LResult.iType));
            end
            else
              raise TCompilerException.Create(CErrorEmptyReturnExpression);
          end;
        end;
      end;
    end;
  end;
var
  LErrorText              : String;
  LFile                   : TStringList;
  LFileInfos              : TFileInfoArray;
  LFloatHelper            : LongInt;
  LFuncHeaderInfo         : TFuncHeaderInfo;
  LFuncHeaderInfos        : TFuncHeaderInfoArray;
  LFuncHeaderInfosChecked : TFuncHeaderInfoArray;
  LIndexA                 : LongInt;
  LIndexB                 : LongInt;
  LIndexC                 : LongInt;
  LInputSource            : String;
  LIntHelper              : LongInt;
  LLineCount              : LongInt;
  LLineIndex              : LongInt;
  LMainEntryPoint         : LongInt;
  LNewOutputSource        : String;
  LOutputSource           : String;
  LResult                 : TInterValue;
  LSymbolNr               : LongInt;
  LSymbols                : TSymbolArray;
  LSymbolTable            : TSymbolTable;
  LTempExpression         : TExpInfo;
  LVarInfos               : TVarInfoArray;
begin
  WriteLn('Progress:');

  SetLength(LFileInfos, 0);
  SetLength(LFuncHeaderInfosChecked, 0);
  SetLength(LSymbols, 0);
  SetLength(LVarInfos, 0);

  LFile := TStringList.Create;
  try
    LFile.LoadFromFile(AInputFileName);
    LInputSource := LFile.Text;

    LSymbolTable := TSymbolTable.Create(ChangeFileExt(ExtractFileName(AInputFileName), ''));
    try
      WriteLn('Killing Comments');
      WriteLn('Killing Compiler Switches');
      WriteLn('Splitting Source File to Chunks');
      LFileInfos := ParseFileParts(KillSwitches(KillQuotes(LInputSource)));

      WriteLn('Processing Chunks');
      SetLength(LFuncHeaderInfos, 0);
      for LIndexA := 0 to Pred(Length(LFileInfos)) do
      begin
        case LFileInfos[LIndexA].fType of
          fitForward, fitFunction :
          begin
            Write('  Parsing Function Definition: ');
            LFuncHeaderInfo := ParseFuncHeader(LFileInfos[LIndexA].fSource);
            WriteLn(LFuncHeaderInfo.fName);

            SetLength(LFuncHeaderInfos, Succ(Length(LFuncHeaderInfos)));
            LFuncHeaderInfos[High(LFuncHeaderInfos)] := LFuncHeaderInfo;
          end;

          fitVariable :
          begin
            Write('  Parsing Variable Definition: ');
            LVarInfos := ParseVarDef(LFileInfos[LIndexA].fSource);
            for LIndexB := 0 to Pred(Length(LVarInfos)) do
            begin
              Write(LVarInfos[LIndexB].vName + ' ');
              if (LVarInfos[LIndexB].vType <> sttNone) then
                LSymbolTable.AddEntry(LVarInfos[LIndexB].vName,
                                      LVarInfos[LIndexB].vType,
                                      sttNone, 0, 0, 0)
              else
                raise TCompilerException.Create(CErrorVoidVariable + LVarInfos[LIndexB].vName);
            end;
            WriteLn('');
            WriteLn('  Variables Written to Symbol Table');
          end;
        end;
      end;

      WriteLn('Checking Function Definitions');
      LFuncHeaderInfosChecked := CheckFuncHeaders(LFuncHeaderInfos);
      for LIndexB := 0 to Pred(Length(LFuncHeaderInfosChecked)) do
      begin
        WriteLn('  Writing Function to Symbol Table: ' + LFuncHeaderInfosChecked[LIndexB].fName);
        LSymbolNr := LSymbolTable.AddEntry(LFuncHeaderInfosChecked[LIndexB].fName,
                                           sttNone,
                                           LFuncHeaderInfosChecked[LIndexB].fType,
                                           0, 0, Length(LFuncHeaderInfosChecked[LIndexB].fParams));

        Write('  Writing Parameters to Symbol Table: ');
        for LIndexC := 0 to Pred(Length(LFuncHeaderInfosChecked[LIndexB].fParams)) do
        begin
          Write(LFuncHeaderInfosChecked[LIndexB].fParams[LIndexC].pName + ' ');
          if (GetVariableType(LFuncHeaderInfosChecked[LIndexB].fParams[LIndexC].pType) <> sttNone) then
            LSymbolTable.AddEntry(LFuncHeaderInfosChecked[LIndexB].fParams[LIndexC].pName,
                                  GetVariableType(LFuncHeaderInfosChecked[LIndexB].fParams[LIndexC].pType),
                                  sttNone, 0, LSymbolNr, Succ(LIndexC))
          else
            raise TCompilerException.Create(CErrorVoidVariable + LFuncHeaderInfosChecked[LIndexB].fParams[LIndexC].pName);
        end;
        WriteLn('');
      end;

      LLineCount      := 0;
      LLineIndex      := 1; // spare line 0 for goto to program
      LMainEntryPoint := 0;
      LOutputSource   := '';
      for LIndexA := 0 to Pred(Length(LFuncHeaderInfos)) do
      begin
        WriteLn('Compiling Function Definition: ' + LFuncHeaderInfos[LIndexA].fName);
        if (LFuncHeaderInfos[LIndexA].fStatement.sType <> sitNone) then
        begin
          if (LFuncHeaderInfos[LIndexA].fName = CCMain) then
          begin
            if (Length(LFuncHeaderInfos[LIndexA].fParams) = 0) then
            begin
              if (LFuncHeaderInfos[LIndexA].fType = sttInt) then
              begin
                WriteLn('  Saving Main Function Position');
                LMainEntryPoint := LLineIndex;
              end
              else
                raise TCompilerException.Create(CErrorNoIntMainFunction);
            end
            else
            begin
              LErrorText := LFuncHeaderInfos[LIndexA].fParams[0].pName;
              for LIndexB := 1 to Pred(Length(LFuncHeaderInfos[LIndexA].fParams)) do
                LErrorText := LErrorText + ', ' + LFuncHeaderInfos[LIndexA].fParams[LIndexB].pName;

              raise TCompilerException.Create(CErrorUnsupportedParametersForMain + LErrorText);
            end;
          end;

          LFloatHelper := 0;
          LIntHelper   := 0;
          LSymbolNr    := LSymbolTable.EntryExists(LFuncHeaderInfos[LIndexA].fName, 0);
          if (LSymbolNr >= 0) then
          begin
            WriteLn('  Setting Function Position in Symbol Table');
            TSymbolEntry(LSymbolTable.Items[LSymbolNr]).Line := LLineIndex;
            WriteLn('  Compiling Source');
            CompileStatement(@LFuncHeaderInfos[LIndexA].fStatement, LLineIndex,
                             LOutputSource, LSymbolTable, LSymbolNr, LFuncHeaderInfos,
                             LIndexA, LFloatHelper, LIntHelper);

            WriteLn('  Writing Mandatory Return');
            if (LFuncHeaderInfos[LIndexA].fType = sttNone) then
            begin
              LOutputSource := LOutputSource + IntToStr(LLineIndex) + #9 + 'RETURN' + sLineBreak;
              Inc(LLineIndex);
            end
            else
            begin
              LOutputSource := LOutputSource + IntToStr(LLineIndex) + #9 + 'RETURN 0' + sLineBreak;
              Inc(LLineIndex);
            end;
          end
          else
            raise TCompilerException.Create(CErrorUnknownSymbol + LFuncHeaderInfos[LIndexA].fName);
        end;
      end;
      WriteLn('Writing Jump to Program Start');
      LOutputSource := '0' + #9 + 'GOTO ' + IntToStr(LLineIndex) + sLineBreak + LOutputSource;

      for LIndexA := 0 to Pred(Length(LFileInfos)) do
      begin
        case LFileInfos[LIndexA].fType of
          fitVariable :
          begin
            LVarInfos := ParseVarDef(LFileInfos[LIndexA].fSource);
            for LIndexB := 0 to Pred(Length(LVarInfos)) do
            begin
              WriteLn('Compiling Variable Statement: ' + LVarInfos[LIndexB].vName);
              if (LVarInfos[LIndexB].vType <> sttNone) then
              begin
                WriteLn('  Checking if Using Variables or Function Calls');
                LSymbols := GetSymbols(LVarInfos[LIndexB].vValue);
                if (Length(LSymbols) > 0) then
                  raise TCompilerException.Create(CErrorGlobalVariableWithoutConstantValue + LVarInfos[LIndexB].vName);

                WriteLn('  Generating Source');
                LResult := GenerateExpression(LVarInfos[LIndexB].vValue,
                                              LLineIndex, LNewOutputSource,
                                              LLineCount, LSymbolTable, 0,
                                              LFloatHelper, LIntHelper);
                LLineIndex    := LLineIndex + LLineCount;
                LOutputSource := LOutputSource + LNewOutputSource;

                WriteLn('  Writing Assignment');
                LTempExpression := ParseExpression(LVarInfos[LIndexB].vName + ' = ' + LResult.iValue);
                WriteLn('  Generating Assignment Source');
                GenerateExpression(LTempExpression, LLineIndex, LNewOutputSource,
                                   LLineCount, LSymbolTable, 0, LFloatHelper, LIntHelper);
                LLineIndex    := LLineIndex + LLineCount;
                LOutputSource := LOutputSource + LNewOutputSource;
              end
              else
                raise TCompilerException.Create(CErrorVoidVariable + LVarInfos[LIndexB].vName);
            end;
          end;
        end;
      end;

      if (LMainEntryPoint > 0) then
      begin
        WriteLn('Writing Main Helper to Symbol Table: ' + CIntHelper + '0');
        LSymbolTable.AddEntry(CIntHelper + '0', sttInt, sttNone, 0, 0, 0);
        WriteLn('Writing Call to Main');
        LOutputSource := LOutputSource + IntToStr(LLineIndex) + #9 + CIntHelper + '0 := CALL ' + CCMain + ', 0' + sLineBreak;
        Inc(LLineIndex);
        WriteLn('Writing Program Return');
        LOutputSource := LOutputSource + IntToStr(LLineIndex) + #9 + 'RETURN';
        Inc(LLineIndex);
      end
      else
        raise TCompilerException.Create(CErrorNoMainFunction);

      WriteLn('');
      WriteLn('Compiled Source:');
      WriteLn(LSymbolTable.PrintSymbolTable);
      WriteLn(LOutputSource);
    except
      on LException : Exception do
      begin
        WriteLn('');
        WriteLn(LException.ClassName);
        WriteLn(LException.Message);
      end;
    end;

    LFile.Text := LSymbolTable.PrintSymbolTable + sLineBreak + LOutputSource;
    LFile.SaveToFile(AOutputFileName);
  finally
    LFile.Free;
//      LFile := nil;
  end;
end;

end.
