unit ExpU;

(*
//  C-to-Quadcode Compiler
//  ----------------------
//  Practical Work:
//      Compilers Summer Semester 2008
//      (Lecture by Mr. Schwinn)
//
//  Course:
//      TAI-06-ABC
//
//  Developers:
//     Kevin Niehage   (Matrikelnummer: 280888)
//     Michaela Hansel (Matrikelnummer: 221664)
*)

// IMPLEMENTATION STATUS: DONE
{.$DEFINE DEBUG_MODE}

interface

uses
  SysUtils,
  ConstsU,
  ErrorsU,
  TypesU;

procedure FreeExpInfo(var AExpInfo : TExpInfo);
function GetSymbols(const AExpInfo : TExpInfo) : TSymbolArray;
function OperationToString(const AOperation : TExpOperation) : String;

function ParseExpression(const AExpression : String) : TExpInfo;
function ParseFuncCall(const AFuncCall : String) : TFuncCallInfo;

implementation

procedure FreeExpInfo(var AExpInfo : TExpInfo);
  procedure FreeExpEntry(var AExpEntry : PExpEntry);
  begin
    if (AExpEntry <> nil) then
    begin
      if (AExpEntry^.eLeft <> nil) then
        FreeExpEntry(AExpEntry^.eLeft);

      if (AExpEntry^.eRight <> nil) then
        FreeExpEntry(AExpEntry^.eRight);

      Dispose(AExpEntry);
      AExpEntry := nil;
    end;
  end;
begin
  FreeExpEntry(AExpInfo.eEntry);
end;

function GetSymbols(const AExpInfo : TExpInfo) : TSymbolArray;
  procedure MergeArrays(var AResult : TSymbolArray; const ANew : TSymbolArray);
  var
    LCount  : LongInt;
    LFound  : LongInt;
    LIndexA : LongInt;
    LIndexB : LongInt;
  begin
    SetLength(AResult, Length(AResult) + Length(ANew));
    LCount := 0;

    for LIndexA := 0 to Pred(Length(ANew)) do
    begin
      LFound := -1;
      for LIndexB := 0 to Pred(Length(AResult)) do
      begin
        if ((ANew[LIndexA].sName = AResult[LIndexB].sName) and
            (ANew[LIndexA].sType = AResult[LIndexB].sType)) then
        begin
          LFound := LIndexB;

          Break;
        end;
      end;

      if (LFound < 0) then
      begin
        AResult[Length(AResult) - Length(ANew) + LCount] := ANew[LIndexA];
        Inc(LCount);
      end;
    end;

    SetLength(AResult, Length(AResult) - Length(ANew) + LCount);
  end;

  function GetExpSymbols(const AExpEntry : PExpEntry) : TSymbolArray;
  var
    LCount   : LongInt;
    LIndexA  : LongInt;
    LIndexB  : LongInt;
    LSymbols : TSymbolArray;
  begin
    SetLength(Result, 0);

    if (AExpEntry <> nil) then
    begin
      SetLength(Result, Length(AExpEntry^.eValues));
      SetLength(LSymbols, 0);
      LCount := 0;
      for LIndexA := 0 to Pred(Length(AExpEntry^.eValues)) do
      begin
        if (AExpEntry^.eValues[LIndexA].vType in [evtFunctionCall, evtVariable]) then
        begin
          if (AExpEntry^.eValues[LIndexA].vType = evtFunctionCall) then
          begin
            Result[LCount].sType := stFunction;

            for LIndexB := 0 to Pred(Length(AExpEntry^.eValues[LIndexA].vParams)) do
              MergeArrays(LSymbols, GetExpSymbols(AExpEntry^.eValues[LIndexA].vParams[LIndexB].eEntry));
          end
          else
            Result[LCount].sType := stVariable;

          Result[LCount].sName := AExpEntry^.eValues[LIndexA].vValue;
          Inc(LCount);
        end;
      end;
      SetLength(Result, LCount);
      MergeArrays(Result, LSymbols);

      if (AExpEntry^.eLeft <> nil) then
        MergeArrays(Result, GetExpSymbols(AExpEntry^.eLeft));

      if (AExpEntry^.eRight <> nil) then
        MergeArrays(Result, GetExpSymbols(AExpEntry^.eRight));
    end;
  end;
begin
  Result := GetExpSymbols(AExpInfo.eEntry);
end;

function OperationToString(const AOperation : TExpOperation) : String;
begin
  Result := '';

  case AOperation of
    eoNone            : Result := 'NONE';
    eoPostDecVal      : Result := 'x--';
    eoPostIncVal      : Result := 'x++';
    eoNotLog          : Result := '!';
    eoNotVal          : Result := '~';
    eoPreDecVal       : Result := '--x';
    eoPreIncVal       : Result := '++x';
    eoUnaryMinusVal   : Result := '-x';
    eoUnaryPlusVal    : Result := '+x';
    eoDivVal          : Result := '/';
    eoModVal          : Result := '%';
    eoMulVal          : Result := '*';
    eoMinusVal        : Result := '-';
    eoPlusVal         : Result := '+';
    eoShlVal          : Result := '<<';
    eoShrVal          : Result := '>>';
    eoGreaterEqualLog : Result := '>=';
    eoGreaterLog      : Result := '>';
    eoLessEqualLog    : Result := '<=';
    eoLessLog         : Result := '<';
    eoEqualLog        : Result := '==';
    eoNotEqualLog     : Result := '!=';
    eoAndVal          : Result := '&';
    eoXorVal          : Result := '^';
    eoOrVal           : Result := '|';
    eoAndLog          : Result := '&&';
    eoOrLog           : Result := '||';
    {eoAndAssign       : Result := '&=';}
    eoAssign          : Result := '=';
    {eoDivAssign       : Result := '/=';
    eoMinusAssign     : Result := '-=';
    eoModAssign       : Result := '%=';
    eoMulAssign       : Result := '*=';
    eoOrAssign        : Result := '|=';
    eoPlusAssign      : Result := '+=';
    eoShlAssign       : Result := '<<=';
    eoShrAssign       : Result := '>>=';
    eoXorAssign       : Result := '^=';}
  end;
end;

function ParseExpression(const AExpression : String) : TExpInfo;
{$IF DEFINED(DEBUG_MODE)}
  function OptionToString(const AOption : TExpValueOption) : String;
  begin
    Result := '';

    case AOption of
      evoNone    : Result := 'NONE';
      evoPreInc  : Result := '++x';
      evoPreDec  : Result := '--x';
      evoPostInc : Result := 'x++';
      evoPostDec : Result := 'x--';
    end;
  end;

  function TypeToString(const AType : TExpValueType) : String;
  begin
    Result := '';

    case AType of
      evtNone         : Result := 'NONE';
      evtHex          : Result := 'Hexadecimal';
      evtOct          : Result := 'Octal';
      evtInteger      : Result := 'Integer';
      evtFloat        : Result := 'Float';
      evtFunctionCall : Result := 'Function-Call';
      evtVariable     : Result := 'Variable';                          
    end;
  end;
{$IFEND DEFINED(DEBUG_MODE)}

  function GetOperationFromString(const AString : String;
                                  const APreOperand : Boolean = false) : TExpOperation;
  var
    LChar : Char;
  begin
    Result := eoNone;

    if (Length(AString) = 1) then
    begin
      LChar := AString[1];
      case LChar of
        '!' : Result := eoNotLog;
        '~' : Result := eoNotVal;

        '/' : Result := eoDivVal;
        '%' : Result := eoModVal;
        '*' : Result := eoMulVal;

        '-' :
        begin
          if APreOperand then
            Result := eoUnaryMinusVal
          else
            Result := eoMinusVal;
        end;
        '+' :
        begin
          if APreOperand then
            Result := eoUnaryPlusVal
          else
            Result := eoPlusVal;
        end;

        '>' : Result := eoGreaterLog;
        '<' : Result := eoLessLog;

        '&' : Result := eoAndVal;

        '^' : Result := eoXorVal;

        '|' : Result := eoOrVal;

        '=' : Result := eoAssign;
      end;
    end
    else
    begin
      if (AString = '--') then
      begin
        if not(APreOperand) then
          Result := eoPostDecVal
        else
          Result := eoPreDecVal;
      end;
      if (AString = '++') then
      begin
        if not(APreOperand) then
          Result := eoPostIncVal
        else
          Result := eoPreIncVal;
      end;

      if (AString = '<<') then
        Result := eoShlVal;
      if (AString = '>>') then
        Result := eoShrVal;

      if (AString = '>=') then
        Result := eoGreaterEqualLog;
      if (AString = '<=') then
        Result := eoLessEqualLog;

      if (AString = '==') then
        Result := eoEqualLog;
      if (AString = '!=') then
        Result := eoNotEqualLog;

      if (AString = '&&') then
        Result := eoAndLog;

      if (AString = '||') then
        Result := eoOrLog;

      {if (AString = '&=') then
        Result := eoAndAssign;
      if (AString = '/=') then
        Result := eoDivAssign;
      if (AString = '-=') then
        Result := eoMinusAssign;
      if (AString = '%=') then
        Result := eoModAssign;
      if (AString = '*=') then
        Result := eoMulAssign;
      if (AString = '|=') then
        Result := eoOrAssign;
      if (AString = '+=') then
        Result := eoPlusAssign;
      if (AString = '<<=') then
        Result := eoShlAssign;
      if (AString = '>>=') then
        Result := eoShrAssign;
      if (AString = '^=') then
        Result := eoXorAssign;}
    end;
  end;

  function GetOperationPriority(const AOperation : TExpOperation;
                                const ABracketLevel : Byte = 0) : Extended;
  begin
    Result := ABracketLevel;

    case AOperation of
      eoPostDecVal,
      eoPostIncVal :
      begin
        Result := Result + 0.95;
      end;

      eoNotLog,
      eoNotVal,
      eoPreDecVal,
      eoPreIncVal,
      eoUnaryMinusVal,
      eoUnaryPlusVal :
      begin
        Result := Result + 0.90;
      end;

      eoDivVal,
      eoModVal,
      eoMulVal :
      begin
        Result := Result + 0.85;
      end;

      eoMinusVal,
      eoPlusVal :
      begin
        Result := Result + 0.80;
      end;

      eoShlVal,
      eoShrVal :
      begin
        Result := Result + 0.75;
      end;

      eoGreaterEqualLog,
      eoGreaterLog,
      eoLessEqualLog,
      eoLessLog :
      begin
        Result := Result + 0.70;
      end;

      eoEqualLog,
      eoNotEqualLog :
      begin
        Result := Result + 0.65;
      end;

      eoAndVal :
      begin
        Result := Result + 0.60;
      end;

      eoXorVal :
      begin
        Result := Result + 0.55;
      end;

      eoOrVal :
      begin
        Result := Result + 0.50;
      end;

      eoAndLog :
      begin
        Result := Result + 0.45;
      end;

      eoOrLog :
      begin
        Result := Result + 0.40;
      end;

      {eoAndAssign,}
      eoAssign{,
      eoDivAssign,
      eoMinusAssign,
      eoModAssign,
      eoMulAssign,
      eoOrAssign,
      eoPlusAssign,
      eoShlAssign,
      eoShrAssign,
      eoXorAssign} :
      begin
        Result := Result + 0.35;
      end;
    end;
  end;

  function IsRightAssociative(const AOperation : TExpOperation) : Boolean;
  begin
    Result := (AOperation in [eoNotLog, eoNotVal,
                              {eoAndAssign,} eoAssign{, eoDivAssign, eoMinusAssign,
                              eoModAssign, eoMulAssign, eoOrAssign, eoPlusAssign,
                              eoShlAssign, eoShrAssign, eoXorAssign}]);
  end;

  function GetEntryPriority(const AEntry : PExpEntry) : Extended;
  begin
    Result := AEntry^.eBrackets;

    if (AEntry <> nil) then
      Result := Result + GetOperationPriority(AEntry^.eOperation);
  end;

  function SelectOperation(const AOld : String; const ANew : String;
                           const APreOperand : Boolean = false) : String;
  begin
    Result := '';

    if (GetOperationFromString(ANew, APreOperand) <> eoNone) then
      Result := ANew
    else
    begin
      if (GetOperationFromString(AOld, APreOperand) <> eoNone) then
        Result := AOld;
    end;
  end;

  function InsertEntry(const ACurrentRoot : PExpEntry; const AEntry : PExpEntry;
                       const AInsertRightOnEqualPriority : Boolean) : PExpEntry;
  var
    LCurrent     : PExpEntry;
    LInsertRight : Boolean;
    LMostLeft    : Boolean;
  begin
    Result := nil;

    if (AEntry <> nil) then
    begin
      LCurrent  := ACurrentRoot;
      LMostLeft := true;
      repeat
        if (LCurrent <> nil) then
        begin
          LInsertRight := (GetEntryPriority(LCurrent) < GetEntryPriority(AEntry));
          if not(LInsertRight) then
            LInsertRight := (GetEntryPriority(LCurrent) = GetEntryPriority(AEntry)) and
                            AInsertRightOnEqualPriority;

          if not(LInsertRight) then
          begin
            if (LCurrent^.eLeft <> nil) then
              LCurrent := LCurrent^.eLeft
            else
            begin
{$IF DEFINED(DEBUG_MODE)}
              WriteLn('INSERTED LEFT (' + IntToHex(Integer(LCurrent), 8) + '->' +
                      IntToHex(Integer(AEntry), 8) + ')');
{$IFEND DEFINED(DEBUG_MODE)}

              LCurrent^.eLeft := AEntry;

              if LMostLeft then
                Result := AEntry
              else
                Result := ACurrentRoot;
            end;
          end
          else
          begin
            LMostLeft := false;

            if (LCurrent^.eRight <> nil) then
              LCurrent := LCurrent^.eRight
            else
            begin
{$IF DEFINED(DEBUG_MODE)}
              WriteLn('INSERTED RIGHT (' + IntToHex(Integer(LCurrent), 8) + '->' +
                      IntToHex(Integer(AEntry), 8) + ')');
{$IFEND DEFINED(DEBUG_MODE)}

              LCurrent^.eRight := AEntry;

              if (GetEntryPriority(LCurrent) = GetOperationPriority(eoAssign)) then
                Result := AEntry
              else
                Result := ACurrentRoot;
            end;
          end;
        end
        else
          Result := AEntry;
      until (Result <> nil);
    end;
  end;

  procedure AddValue(const AEntry : PExpEntry; const AValue : String;
                     const AType : TExpValueType; const AOption : TExpValueOption);
  var
    LFuncCall : TFuncCallInfo;
    LIndex    : LongInt;
  begin
    if (AEntry <> nil) then
    begin
      SetLength(AEntry^.eValues, Succ(Length(AEntry^.eValues)));
      AEntry^.eValues[High(AEntry^.eValues)].vOption := AOption;
      AEntry^.eValues[High(AEntry^.eValues)].vType   := AType;

      if (AType <> evtFunctionCall) then
      begin
{$IF DEFINED(DEBUG_MODE)}
        WriteLn('[ADD] Value: ' + AValue + ' Type: ' + TypeToString(AType) +
                ' Option: ' + OptionToString(AOption) + ' (' + IntToHex(Integer(AEntry), 8) + ')');
{$IFEND DEFINED(DEBUG_MODE)}

        AEntry^.eValues[High(AEntry^.eValues)].vParams := nil;
        AEntry^.eValues[High(AEntry^.eValues)].vValue  := AValue;
      end
      else
      begin
        LFuncCall := ParseFuncCall(AValue);

{$IF DEFINED(DEBUG_MODE)}
        WriteLn('[ADD] Value: ' + LFuncCall.fName + ' Type: ' + TypeToString(AType) +
                ' Option: ' + OptionToString(AOption) + ' (' + IntToHex(Integer(AEntry), 8) + ')');
{$IFEND DEFINED(DEBUG_MODE)}

        AEntry^.eValues[High(AEntry^.eValues)].vValue := LFuncCall.fName;

        SetLength(AEntry^.eValues[High(AEntry^.eValues)].vParams, Length(LFuncCall.fParams));
        for LIndex := 0 to Pred(Length(LFuncCall.fParams)) do
        begin
{$IF DEFINED(DEBUG_MODE)}
          WriteLn('[PARSE] Value: ' + LFuncCall.fParams[LIndex] + ' (' + IntToHex(Integer(AEntry), 8) + ')');
{$IFEND DEFINED(DEBUG_MODE)}

          AEntry^.eValues[High(AEntry^.eValues)].vParams[LIndex] := ParseExpression(LFuncCall.fParams[LIndex]);
        end;
      end;
    end;
  end;

  function NewEntry(const AOperation : TExpOperation; const ABracketLevel : LongInt) : PExpEntry;
  begin
    New(Result);

    Result^.eBrackets  := ABracketLevel;
    Result^.eLeft      := nil;
    Result^.eOperation := AOperation;
    Result^.eRight     := nil;
    SetLength(Result^.eValues, 0);

{$IF DEFINED(DEBUG_MODE)}
    WriteLn('[NEW] Operation: ' + OperationToString(AOperation) +
            ' BracketLevel: ' + IntToStr(ABracketLevel) +
            ' Priority: ' + FloatToStr(GetOperationPriority(AOperation, ABracketLevel)) +
            ' (' + IntToHex(Integer(Result), 8) + ')');
{$IFEND DEFINED(DEBUG_MODE)}
  end;

var
  LBracketCount : LongInt;
  LBrackets     : LongInt;
  LBuffer       : String;
  LCurrentRoot  : PExpEntry;
  LEntry        : PExpEntry;
  LIndex        : LongInt;
  LNegative     : Boolean;
  LNext         : Char;
  LOption       : TExpValueOption;
  LStatus       : TExpStatus;
  LTempEntry    : PExpEntry;
  LTempOp       : TExpOperation;
  LTempPrio     : Extended;
  LType         : TExpValueType;
  LValBuffer    : String;
begin
  Result.eEntry := nil;

  LBracketCount := 0;
  LBrackets     := 0;
  LBuffer       := '';
  LCurrentRoot  := nil;
  LEntry        := nil;
  LNegative     := false;
  LNext         := #0;
  LOption       := evoNone;
  LStatus       := esWaitBefore;
//  LTempEntry    := nil;
//  LTempOp       := eoNone;
//  LTempPrio     := 0.0;
  LType         := evtNone;
  LValBuffer    := '';

  LIndex := 1;
  while (LIndex <= Length(AExpression)) do
  begin
    LNext := AExpression[LIndex];

    case LStatus of
      esWaitBefore :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            // do nothing
          end;

          'A'..'Z', 'a'..'z', '_' : // variable
          begin
            LBuffer := LBuffer + LNext;
            LType   := evtVariable;

            LStatus := esVariable;
          end;

          '0' : // oct or hex
          begin
            LBuffer := LBuffer + LNext;
            LType   := evtInteger;

            LStatus := esHexOrOct;
          end;

          '1'..'9' : // integer
          begin
            LBuffer := LBuffer + LNext;
            LType   := evtInteger;

            LStatus := esInteger;
          end;

          '.' : // float
          begin
            LBuffer := LBuffer + LNext;
            LType   := evtFloat;

            LStatus := esFloat;
          end;

          '(' : // higher priority
          begin
            Inc(LBrackets);
          end;
        else
          if (GetOperationFromString(LNext, (Length(LBuffer) <= 0)) <> eoNone) then
          begin
            LStatus := esOperator;
            Dec(LIndex); // reset lookahead
          end
          else
            raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorUnknownCharacter));
        end;
      end;

      esHexOrOct :
      begin
        case LNext of
          '0'..'7' : // oct
          begin
            LBuffer := LBuffer + LNext;
            LType   := evtOct;

            LStatus := esOct;
          end;

          'x' : // hex
          begin
            LType := evtHex;

            LStatus := esHexBeforeNum;
          end;

          ')' : // lower priority
          begin
            if (LBrackets > 0) then
              Dec(LBrackets)
            else
              raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorBracketCountMismatch));

            LValBuffer := LBuffer;
            LBuffer    := '';

            LStatus := esOperator;
          end;
        else
          if (GetOperationFromString(LNext, (Length(LBuffer) <= 0)) <> eoNone) then
          begin
            LValBuffer := LBuffer;
            LBuffer    := '';

            LStatus := esOperator;
            Dec(LIndex); // reset lookahead
          end
          else
            raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorUnknownCharacter));
        end;
      end;

      esHexBeforeNum :
      begin
        case LNext of
          '0'..'9', 'A'..'F', 'a'..'f' : // hex
          begin
            LBuffer := LBuffer + LNext;

            LStatus := esHex;
          end;
        else
          raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorUnknownCharacter));
        end;
      end;

      esHex :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            LValBuffer := LBuffer;
            LBuffer    := '';

            LStatus := esOperator;
          end;

          '0'..'9', 'A'..'F', 'a'..'f' : // hex
          begin
            LBuffer := LBuffer + LNext;
          end;

          ')' : // lower priority
          begin
            if (LBrackets > 0) then
              Dec(LBrackets)
            else
              raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorBracketCountMismatch));

            LValBuffer := LBuffer;
            LBuffer    := '';

            LStatus := esOperator;
          end;
        else
          if (GetOperationFromString(LNext, (Length(LBuffer) <= 0)) <> eoNone) then
          begin
            LValBuffer := LBuffer;
            LBuffer    := '';

            LStatus := esOperator;
            Dec(LIndex); // reset lookahead
          end
          else
            raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorUnknownCharacter));
        end;
      end;

      esOct :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            LValBuffer := LBuffer;
            LBuffer    := '';

            LStatus := esOperator;
          end;

          '0'..'7' : // oct
          begin
            LBuffer := LBuffer + LNext;
          end;

          ')' : // lower priority
          begin
            if (LBrackets > 0) then
              Dec(LBrackets)
            else
              raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorBracketCountMismatch));

            LValBuffer := LBuffer;
            LBuffer    := '';

            LStatus := esOperator;
          end;
        else
          if (GetOperationFromString(LNext, (Length(LBuffer) <= 0)) <> eoNone) then
          begin
            LValBuffer := LBuffer;
            LBuffer    := '';

            LStatus := esOperator;
            Dec(LIndex); // reset lookahead
          end
          else
            raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorUnknownCharacter));
        end;
      end;

      esInteger :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            LValBuffer := LBuffer;
            LBuffer    := '';

            LStatus := esOperator;
          end;

          '0'..'9' : // integer
          begin
            LBuffer := LBuffer + LNext;
          end;

          '.' : // float
          begin
            LBuffer := LBuffer + LNext;
            LType   := evtFloat;

            LStatus := esFloat;
          end;

          'E', 'e' : // exponent
          begin
            LBuffer := LBuffer + LNext;
            LType   := evtFloat;

            LStatus := esExponentBeforeSign;
          end;

          ')' : // lower priority
          begin
            if (LBrackets > 0) then
              Dec(LBrackets)
            else
              raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorBracketCountMismatch));

            LValBuffer := LBuffer;
            LBuffer    := '';

            LStatus := esOperator;
          end;
        else
          if (GetOperationFromString(LNext, (Length(LBuffer) <= 0)) <> eoNone) then
          begin
            LValBuffer := LBuffer;
            LBuffer    := '';

            LStatus := esOperator;
            Dec(LIndex); // reset lookahead
          end
          else
            raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorUnknownCharacter));
        end;
      end;

      esFloat :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            LValBuffer := LBuffer;
            LBuffer    := '';

            LStatus := esOperator;
          end;

          'E', 'e' : // exponent
          begin
            LBuffer := LBuffer + LNext;
            LType   := evtFloat;

            LStatus := esExponentBeforeSign;
          end;

          '0'..'9' : // integer
          begin
            LBuffer := LBuffer + LNext;
          end;

          ')' : // lower priority
          begin
            if (LBrackets > 0) then
              Dec(LBrackets)
            else
              raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorBracketCountMismatch));

            LValBuffer := LBuffer;
            LBuffer    := '';

            LStatus := esOperator;
          end;
        else
          if (GetOperationFromString(LNext, (Length(LBuffer) <= 0)) <> eoNone) then
          begin
            LValBuffer := LBuffer;
            LBuffer    := '';

            LStatus := esOperator;
            Dec(LIndex); // reset lookahead
          end
          else
            raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorUnknownCharacter));
        end;
      end;

      esExponentBeforeSign :
      begin
        case LNext of
          '0'..'9' : // exponent
          begin
            LBuffer := LBuffer + LNext;

            LStatus := esExponent;
          end;

          '+', '-' : // sign
          begin
            LBuffer := LBuffer + LNext;

            LStatus := esExponentBeforeNum;
          end;
        else
          raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorUnknownCharacter));
        end;
      end;

      esExponentBeforeNum :
      begin
        case LNext of
          '0'..'9' : // exponent
          begin
            LBuffer := LBuffer + LNext;

            LStatus := esExponent;
          end;
        else
          raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorUnknownCharacter));
        end;
      end;

      esExponent :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            LValBuffer := LBuffer;
            LBuffer    := '';

            LStatus := esOperator;
          end;

          '0'..'9' : // exponent
          begin
            LBuffer := LBuffer + LNext;
          end;

          ')' : // lower priority
          begin
            if (LBrackets > 0) then
              Dec(LBrackets)
            else
              raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorBracketCountMismatch));

            LValBuffer := LBuffer;
            LBuffer    := '';

            LStatus := esOperator;
          end;
        else
          if (GetOperationFromString(LNext, (Length(LBuffer) <= 0)) <> eoNone) then
          begin
            LValBuffer := LBuffer;
            LBuffer    := '';

            LStatus := esOperator;
            Dec(LIndex); // reset lookahead
          end
          else
            raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorUnknownCharacter));
        end;
      end;

      esVariable :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
//            LValBuffer := LBuffer;
//            LBuffer    := '';

//            LStatus := esOperator;
          end;

          'A'..'Z', 'a'..'z', '_', '0'..'9' : // variable
          begin
            LBuffer := LBuffer + LNext;
          end;

          '(' : // function call
          begin
            LBracketCount := 1;
            LBuffer       := LBuffer + LNext;
            LType         := evtFunctionCall;

            LStatus := esFunctionCall;
          end;

          ')' : // lower priority
          begin
            if (LBrackets > 0) then
              Dec(LBrackets)
            else
              raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorBracketCountMismatch));

            LValBuffer := LBuffer;
            LBuffer    := '';

            LStatus := esOperator;
          end;
        else
          if (GetOperationFromString(LNext, (Length(LBuffer) <= 0)) <> eoNone) then
          begin
            LValBuffer := LBuffer;
            LBuffer    := '';

            LStatus := esOperator;
            Dec(LIndex); // reset lookahead
          end
          else
            raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorUnknownCharacter));
        end;
      end;

      esFunctionCall :
      begin
        LBuffer := LBuffer + LNext;

        case LNext of
          '(' : // count bracket + 1
          begin
            Inc(LBracketCount);
          end;

          ')' : // count bracket - 1
          begin
            Dec(LBracketCount);
            if (LBracketCount = 0) then
            begin
              LValBuffer := LBuffer;
              LBuffer    := '';

              LStatus := esOperator;
            end;
          end;
        end;
      end;

      esOperator :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            if (Length(LBuffer) > 0) then
              LStatus := esWriteOperator;
          end;

          '+', '-', '*', '/', '%', '&', '|', '^', '=', '<', '>', '!', '~' : // operator
          begin
            if (SelectOperation(LBuffer, LBuffer + LNext, (LType = evtNone)) = LBuffer + LNext) then
              LBuffer := LBuffer + LNext
            else
            begin
              LStatus := esWriteOperator;
              Dec(LIndex);
            end;
          end;

          ')' : // lower priority
          begin
            if (LBrackets > 0) then
              Dec(LBrackets)
            else
              raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorBracketCountMismatch));
          end;
        else
          if (Length(LBuffer) > 0) then
          begin
            LStatus := esWriteOperator;
            Dec(LIndex);
          end
          else
            raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorUnknownCharacter));
        end;
      end;

      esWriteOperator :
      begin
        if (Length(LBuffer) >= 0) then
        begin
          LTempOp := GetOperationFromString(LBuffer, (LType = evtNone));
          if (LTempOp <> eoNone) then
          begin
            if (LTempOp in [eoPostDecVal, eoPostIncVal, eoPreDecVal,
                            eoPreIncVal, eoUnaryMinusVal, eoUnaryPlusVal]) then
            begin // optional operators
              case LTempOp of
                eoPostDecVal :
                begin
                  if (LOption = evoNone) then
                    LOption := evoPostDec
                  else
                    raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorDuplicateIncOperation));
                end;

                eoPostIncVal :
                begin
                  if (LOption = evoNone) then
                    LOption := evoPostInc
                  else
                    raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorDuplicateIncOperation));
                end;

                eoPreDecVal :
                begin
                  if (LOption = evoNone) then
                    LOption := evoPreDec
                  else
                    raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorDuplicateIncOperation));
                end;

                eoPreIncVal :
                begin
                  if (LOption = evoNone) then
                    LOption := evoPreInc
                  else
                    raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorDuplicateIncOperation));
                end;

                eoUnaryMinusVal :
                begin
                  LNegative := not(LNegative);
                end;

                eoUnaryPlusVal :
                begin
                // do nothing
                end;
              end;
            end
            else
            begin
              if (LTempOp in [eoNotLog, eoNotVal]) then
              begin // unary operators
                if (LType = evtNone) then
                begin
                  LEntry       := NewEntry(LTempOp, LBrackets);
                  LCurrentRoot := InsertEntry(LCurrentRoot, LEntry, IsRightAssociative(LEntry^.eOperation));
                end
                else
                  raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorUnaryOperationWithOperand));
              end
              else
              begin // binary operators
                if (LType <> evtNone) then
                begin
                  if LNegative then
                    LValBuffer := '-' + LValBuffer;

                  LTempEntry := NewEntry(LTempOp, LBrackets);
                  if (LEntry = nil) then
                    LEntry := LTempEntry;
                  if (LCurrentRoot = nil) then
                    LCurrentRoot := LEntry;

                  LTempPrio := GetEntryPriority(LEntry) - GetEntryPriority(LTempEntry);
                  if (LTempPrio <> 0.0) then
                  begin
                    if (LTempPrio > 0.0) then
                    begin
                      if IsRightAssociative(LEntry^.eOperation) then
                      begin
                        LEntry := LTempEntry;
                        AddValue(LEntry, LValBuffer, LType, LOption);
                      end
                      else
                      begin
                        AddValue(LEntry, LValBuffer, LType, LOption);
                        LEntry := LTempEntry;
                      end;
                    end
                    else
                    begin
                      LEntry := LTempEntry;
                      AddValue(LEntry, LValBuffer, LType, LOption);
                    end;

                    LCurrentRoot := InsertEntry(LCurrentRoot, LEntry, IsRightAssociative(LEntry^.eOperation));
                  end
                  else
                  begin
                    if (LEntry^.eOperation = LTempEntry^.eOperation) then
                    begin
                      AddValue(LEntry, LValBuffer, LType, LOption);

                      if (LEntry <> LTempEntry) then
                        Dispose(LTempEntry);
                    end
                    else
                    begin
                      if IsRightAssociative(LEntry^.eOperation) then
                      begin
                        LEntry := LTempEntry;
                        AddValue(LEntry, LValBuffer, LType, LOption);
                      end
                      else
                      begin
                        AddValue(LEntry, LValBuffer, LType, LOption);
                        LEntry := LTempEntry;
                      end;

                      LCurrentRoot := InsertEntry(LCurrentRoot, LEntry, IsRightAssociative(LEntry^.eOperation));
                    end;
                  end;

                  LNegative  := false;
                  LOption    := evoNone;
                  LType      := evtNone;
                  LValBuffer := '';
                end
                else
                  raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorBinaryOperationWithoutOperand));
              end;

              if (Result.eEntry = nil) then
              begin
{$IF DEFINED(DEBUG_MODE)}
                WriteLn('INSERTED ROOT (' + IntToHex(Integer(LCurrentRoot), 8) + ')');
{$IFEND DEFINED(DEBUG_MODE)}

                Result.eEntry := LCurrentRoot;
              end;
            end;

            LBuffer := '';
          end
          else
            raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorUnknownOperand));
        end;

        LStatus := esWaitBefore;
        Dec(LIndex);
      end;
    end;

    Inc(LIndex);
  end;

  if ((Length(LValBuffer) > 0) or
      (Length(LBuffer) > 0)) then
  begin
    if (Result.eEntry = nil) then
    begin
      LEntry := NewEntry(eoPlusVal, 0);
      Result.eEntry := LEntry;

{$IF DEFINED(DEBUG_MODE)}
      WriteLn('INSERTED ROOT (' + IntToHex(Integer(LEntry), 8) + ')');
{$IFEND DEFINED(DEBUG_MODE)}
    end;

    if (Length(LValBuffer) > 0) then
    begin
      LTempOp := GetOperationFromString(LBuffer, (LType = evtNone));
      if (LTempOp <> eoNone) then
      begin
        if (LTempOp in [eoPostDecVal, eoPostIncVal, eoPreDecVal,
                        eoPreIncVal, eoUnaryMinusVal, eoUnaryPlusVal]) then
        begin // optional operators
          case LTempOp of
            eoPostDecVal :
            begin
              if (LOption = evoNone) then
                LOption := evoPostDec
              else
                raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorDuplicateIncOperation));
            end;

            eoPostIncVal :
            begin
              if (LOption = evoNone) then
                LOption := evoPostInc
              else
                raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorDuplicateIncOperation));
            end;

            eoPreDecVal :
            begin
              if (LOption = evoNone) then
                LOption := evoPreDec
              else
                raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorDuplicateIncOperation));
            end;

            eoPreIncVal :
            begin
              if (LOption = evoNone) then
                LOption := evoPreInc
              else
                raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorDuplicateIncOperation));
            end;

            eoUnaryPlusVal :
            begin
            // do nothing
            end;
          else
            raise TExpInfoException.Create(CErrorNoEnding + LBuffer);
          end;
        end;
      end;

      if LNegative then
        LValBuffer := '-' + LValBuffer;
      AddValue(LEntry, LValBuffer, LType, LOption);
    end
    else
    begin
      if LNegative then
        LBuffer := '-' + LBuffer;
      AddValue(LEntry, LBuffer, LType, LOption);
    end;
  end;

  if (LBrackets <> 0) then
    raise TExpInfoException.Create(BuildParseError(AExpression, LNext, LBuffer, Length(LBuffer), CErrorBracketCountMismatch));
end;


function ParseFuncCall(const AFuncCall : String) : TFuncCallInfo;
var
  LBracketCount : LongInt;
  LBuffer       : String;
  LCount        : LongInt;
  LIndex        : LongInt;
  LNext         : Char;
  LStatus       : TFuncCallStatus;
begin
  Result.fName := '';
  SetLength(Result.fParams, 0);

  SetLength(LBuffer, Length(AFuncCall));
  LCount  := 0;

  LBracketCount := 0;
  LNext         := #0;
  LStatus       := fcsWaitBeforeName;
  for LIndex := 1 to Length(AFuncCall) do
  begin
    LNext := AFuncCall[LIndex];

    case LStatus of
      fcsWaitBeforeName :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            // do nothing
          end;

          'A'..'Z', 'a'..'z', '_' : // name
          begin
            Inc(LCount);
            LBuffer[LCount] := LNext;

            LStatus := fcsName;
          end;
        else
          raise TFuncCallInfoException.Create(BuildParseError(AFuncCall, LNext, LBuffer, LCount, CErrorUnknownNameCharacter));
        end;
      end;

      fcsName :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            SetLength(LBuffer, LCount);
            Result.fName := LBuffer;

            SetLength(LBuffer, Length(AFuncCall) - LCount);
            LCount := 0;

            LStatus := fcsWaitAfterName;
          end;

          'A'..'Z', 'a'..'z', '_', '0'..'9' : // name
          begin
            Inc(LCount);
            LBuffer[LCount] := LNext;
          end;

          '(' : // begin parameters
          begin
            LBracketCount := 1;

            SetLength(LBuffer, LCount);
            Result.fName := LBuffer;

            SetLength(LBuffer, Length(AFuncCall) - LCount);
            LCount := 0;

            LStatus := fcsParam;
          end;
        else
          raise TFuncCallInfoException.Create(BuildParseError(AFuncCall, LNext, LBuffer, LCount, CErrorUnknownNameCharacter));
        end;
      end;

      fcsWaitAfterName :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            // do nothing
          end;

          '(' : // begin parameters
          begin
            LBracketCount := 1;

            LStatus := fcsParam;
          end;
        else
          raise TFuncCallInfoException.Create(BuildParseError(AFuncCall, LNext, LBuffer, LCount, CErrorUnknownCharacter));
        end;
      end;

      fcsParam :
      begin
        case LNext of
          '(' : // open bracket
          begin
            Inc(LBracketCount);

            Inc(LCount);
            LBuffer[LCount] := LNext;
          end;

          ')' : // close bracket
          begin
            Dec(LBracketCount);

            if (LBracketCount >= 1) then
            begin
              Inc(LCount);
              LBuffer[LCount] := LNext;
            end
            else
            begin
              SetLength(LBuffer, LCount);
              if (LCount > 0) then
              begin
                SetLength(Result.fParams, Succ(Length(Result.fParams)));
                Result.fParams[High(Result.fParams)] := LBuffer;
              end;

              SetLength(LBuffer, Length(AFuncCall) - LCount);
              LCount := 0;

              LStatus := fcsWaitForError;
            end;
          end;

          ',' : // seperator
          begin
            if (LBracketCount = 1) then
            begin
              SetLength(LBuffer, LCount);
              SetLength(Result.fParams, Succ(Length(Result.fParams)));
              Result.fParams[High(Result.fParams)] := LBuffer;

              SetLength(LBuffer, Length(AFuncCall) - LCount);
              LCount := 0;
            end
            else
            begin
              Inc(LCount);
              LBuffer[LCount] := LNext;
            end;
          end;
        else
          Inc(LCount);
          LBuffer[LCount] := LNext;
        end;
      end;

      fcsWaitForError :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            // do nothing
          end;
        else
          raise TFuncCallInfoException.Create(BuildParseError(AFuncCall, LNext, LBuffer, LCount, CErrorUnknownCharacter));
        end;
      end;
    end;
  end;

  if (LStatus <> fcsWaitForError) then
    raise TFuncCallInfoException.Create(BuildParseError(AFuncCall, LNext, LBuffer, LCount, CErrorNoEnding));
end;

end.
