unit FuncsU;

(*
//  C-to-Quadcode Compiler
//  ----------------------
//  Practical Work:
//      Compilers Summer Semester 2008
//      (Lecture by Mr. Schwinn)
//
//  Course:
//      TAI-06-ABC
//
//  Developers:
//     Kevin Niehage   (Matrikelnummer: 280888)
//     Michaela Hansel (Matrikelnummer: 221664)
*)

// IMPLEMENTATION STATUS: DONE

interface

uses
  SysUtils,
  ConstsU,
  ErrorsU,
  StmtsU,
  TypesU,
  VarsU;

function CheckFuncHeaders(const AFuncHeaderList : TFuncHeaderInfoArray) : TFuncHeaderInfoArray;

function ParseFuncHeader(const AFuncHeader : String) : TFuncHeaderInfo;

implementation

function CheckFuncHeaders(const AFuncHeaderList : TFuncHeaderInfoArray) : TFuncHeaderInfoArray;
var
  LCount   : LongInt;
  LDefined : LongInt;
  LEqual   : Boolean;
  LIndexA  : LongInt;
  LIndexB  : LongInt;
  LIndexC  : LongInt;
  LReduced : LongInt;
begin
  SetLength(Result, Length(AFuncHeaderList));
  LCount := 0;

  LReduced := 0;
  for LIndexA := 0 to Pred(Length(AFuncHeaderList)) do
  begin
    if (LIndexA >= (Length(AFuncHeaderList) - LReduced)) then
      Break;

    if (AFuncHeaderList[LIndexA].fStatement.sType = sitNone) then
    begin
      LDefined := LIndexA;

      for LIndexB := Succ(LIndexA) to (Pred(Length(AFuncHeaderList)) - LReduced) do
      begin
        if (Trim(AFuncHeaderList[LIndexA].fName) = Trim(AFuncHeaderList[LIndexB].fName)) then
        begin
          if (AFuncHeaderList[LIndexB].fStatement.sType = sitNone) then
            raise TFuncHeaderInfoException.Create(CErrorDuplicateFunctionForward + AFuncHeaderList[LIndexA].fName)
          else
          begin
            if (LDefined = LIndexA) then
            begin
              LEqual := (AFuncHeaderList[LIndexA].fType = AFuncHeaderList[LIndexB].fType);
              if LEqual then
              begin
                LEqual := (Length(AFuncHeaderList[LIndexA].fParams) = Length(AFuncHeaderList[LIndexB].fParams));
                if LEqual then
                begin
                  for LIndexC := 0 to Pred(Length(AFuncHeaderList[LIndexA].fParams)) do
                  begin
                    LEqual := ((Trim(AFuncHeaderList[LIndexA].fParams[LIndexC].pName) = Trim(AFuncHeaderList[LIndexB].fParams[LIndexC].pName)) and
                               (Trim(AFuncHeaderList[LIndexA].fParams[LIndexC].pType) = Trim(AFuncHeaderList[LIndexB].fParams[LIndexC].pType)));

                    if not(LEqual) then
                      Break;
                  end;
                end;
              end;

              if LEqual then
                LDefined := LIndexB
              else
                raise TFuncHeaderInfoException.Create(CErrorFunctionDefinitionDoesNotMatchForward + AFuncHeaderList[LIndexA].fName);
            end
            else
              raise TFuncHeaderInfoException.Create(CErrorDuplicateFunctionDefinition + AFuncHeaderList[LIndexA].fName);
          end;
        end;
      end;

      if (LDefined >= LIndexA) then
      begin
        Result[LCount] := AFuncHeaderList[LDefined];
        Inc(LCount);

        Inc(LReduced);
        for LIndexC := LDefined to (Pred(Length(AFuncHeaderList)) - LReduced) do
          AFuncHeaderList[LIndexC] := AFuncHeaderList[Succ(LIndexC)];
      end
      else
        raise TFuncHeaderInfoException.Create(CErrorNoFunctionDefinition + AFuncHeaderList[LIndexA].fName);
    end
    else
    begin
      for LIndexB := Succ(LIndexA) to (Pred(Length(AFuncHeaderList)) - LReduced) do
      begin
        if (AFuncHeaderList[LIndexA].fName = AFuncHeaderList[LIndexB].fName) then
        begin
          if (AFuncHeaderList[LIndexB].fStatement.sType = sitNone) then
            raise TFuncHeaderInfoException.Create(CErrorFunctionForwardAfterDefinition + AFuncHeaderList[LIndexA].fName)
          else
            raise TFuncHeaderInfoException.Create(CErrorDuplicateFunctionDefinition + AFuncHeaderList[LIndexA].fName);
        end;
      end;

      Result[LCount] := AFuncHeaderList[LIndexA];
      Inc(LCount);
    end;
  end;

  SetLength(Result, LCount);
end;

function ParseFuncHeader(const AFuncHeader : String) : TFuncHeaderInfo;
var
  LBracketCount : LongInt;
  LBuffer       : String;
  LCount        : LongInt;
  LIndex        : LongInt;
  LNext         : Char;
  LStatus       : TFuncHeaderStatus;
  LType         : String;
begin
  Result.fName    := '';
  Result.fType    := sttNone;
  SetLength(Result.fParams, 0);
  EmptyStatementInfo(Result.fStatement);

  SetLength(LBuffer, Length(AFuncHeader));
  LCount  := 0;

  LBracketCount := 0;
  LNext         := #0;
  LStatus       := fhsWaitBeforeType;
  LType         := '';
  for LIndex := 1 to Length(AFuncHeader) do
  begin
    LNext := AFuncHeader[LIndex];

    case LStatus of
      fhsWaitBeforeType :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            // do nothing
          end;

          'A'..'Z', 'a'..'z' : // type
          begin
            Inc(LCount);
            LBuffer[LCount] := LNext;

            LStatus := fhsType;
          end;
        else
          raise TFuncHeaderInfoException.Create(BuildParseError(AFuncHeader, LNext, LBuffer, LCount, CErrorUnknownTypeCharacter));
        end;
      end;

      fhsType :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            SetLength(LBuffer, LCount);
            Result.fType := GetVariableType(LBuffer);

            SetLength(LBuffer, Length(AFuncHeader) - LCount);
            LCount := 0;

            LStatus := fhsWaitBeforeName;
          end;

          'A'..'Z', 'a'..'z' : // type
          begin
            Inc(LCount);
            LBuffer[LCount] := LNext;
          end;
        else
          raise TFuncHeaderInfoException.Create(BuildParseError(AFuncHeader, LNext, LBuffer, LCount, CErrorUnknownTypeCharacter));
        end;
      end;

      fhsWaitBeforeName :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            // do nothing
          end;

          'A'..'Z', 'a'..'z', '_' : // name
          begin
            Inc(LCount);
            LBuffer[LCount] := LNext;

            LStatus := fhsName;
          end;
        else
          raise TFuncHeaderInfoException.Create(BuildParseError(AFuncHeader, LNext, LBuffer, LCount, CErrorUnknownNameCharacter));
        end;
      end;

      fhsName :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            SetLength(LBuffer, LCount);
            Result.fName := LBuffer;

            SetLength(LBuffer, Length(AFuncHeader) - LCount);
            LCount := 0;

            LStatus := fhsWaitAfterName;
          end;

          'A'..'Z', 'a'..'z', '_', '0'..'9' : // name
          begin
            Inc(LCount);
            LBuffer[LCount] := LNext;
          end;

          '(' : // begin parameters
          begin
            SetLength(LBuffer, LCount);
            Result.fName := LBuffer;

            SetLength(LBuffer, Length(AFuncHeader) - LCount);
            LCount := 0;

            LStatus := fhsWaitBeforeParamType;
          end;
        else
          raise TFuncHeaderInfoException.Create(BuildParseError(AFuncHeader, LNext, LBuffer, LCount, CErrorUnknownNameCharacter));
        end;
      end;

      fhsWaitAfterName :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            // do nothing
          end;

          '(' : // begin parameters
          begin
            LStatus := fhsWaitBeforeParamType;
          end;
        else
          raise TFuncHeaderInfoException.Create(BuildParseError(AFuncHeader, LNext, LBuffer, LCount, CErrorUnknownCharacter));
        end;
      end;

      fhsWaitBeforeParamType :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            // do nothing
          end;

          'A'..'Z', 'a'..'z' : // type
          begin
            Inc(LCount);
            LBuffer[LCount] := LNext;

            LStatus := fhsParamType;
          end;

          ')' : // end parameters
          begin
            LStatus := fhsWaitAfterParam;
          end;
        else
          raise TFuncHeaderInfoException.Create(BuildParseError(AFuncHeader, LNext, LBuffer, LCount, CErrorUnknownTypeCharacter));
        end;
      end;

      fhsParamType :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            SetLength(LBuffer, LCount);
            LType := LBuffer;

            SetLength(LBuffer, Length(AFuncHeader) - LCount);
            LCount := 0;

            LStatus := fhsWaitBeforeParamName;
          end;

          'A'..'Z', 'a'..'z' : // type
          begin
            Inc(LCount);
            LBuffer[LCount] := LNext;
          end;

          ')' : // void
          begin
            SetLength(LBuffer, LCount);
            LBuffer := AnsiLowerCase(Trim(LBuffer));
            if ((LBuffer = AnsiLowerCase(Trim(CCVoid))) and (Length(Result.fParams) = 0)) then
            begin
              SetLength(LBuffer, Length(AFuncHeader) - LCount);
              LCount := 0;

              LStatus := fhsWaitAfterParam;
            end
            else
              raise TFuncHeaderInfoException.Create(BuildParseError(AFuncHeader, LNext, LBuffer, LCount, CErrorUnknownTypeCharacter));
          end;
        else
          raise TFuncHeaderInfoException.Create(BuildParseError(AFuncHeader, LNext, LBuffer, LCount, CErrorUnknownTypeCharacter));
        end;
      end;

      fhsWaitBeforeParamName :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            // do nothing
          end;

          'A'..'Z', 'a'..'z', '_' : // name
          begin
            Inc(LCount);
            LBuffer[LCount] := LNext;

            LStatus := fhsParamName;
          end;

          ')' : // void
          begin
            SetLength(LBuffer, LCount);
            LBuffer := AnsiLowerCase(Trim(LBuffer));
            if ((LBuffer = AnsiLowerCase(Trim(CCVoid))) and (Length(Result.fParams) = 0)) then
            begin
              SetLength(LBuffer, Length(AFuncHeader) - LCount);
              LCount := 0;

              LStatus := fhsWaitAfterParam;
            end
            else
              raise TFuncHeaderInfoException.Create(BuildParseError(AFuncHeader, LNext, LBuffer, LCount, CErrorUnknownTypeCharacter));
          end;
        else
          raise TFuncHeaderInfoException.Create(BuildParseError(AFuncHeader, LNext, LBuffer, LCount, CErrorUnknownNameCharacter));
        end;
      end;

      fhsParamName :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            SetLength(LBuffer, LCount);
            SetLength(Result.fParams, Succ(Length(Result.fParams)));
            Result.fParams[High(Result.fParams)].pName := LBuffer;
            Result.fParams[High(Result.fParams)].pType := LType;

            SetLength(LBuffer, Length(AFuncHeader) - LCount);
            LCount := 0;

            LStatus := fhsWaitAfterParamName;
          end;

          'A'..'Z', 'a'..'z', '_', '0'..'9' : // name
          begin
            Inc(LCount);
            LBuffer[LCount] := LNext;
          end;

          ',' : // type
          begin
            SetLength(LBuffer, LCount);
            SetLength(Result.fParams, Succ(Length(Result.fParams)));
            Result.fParams[High(Result.fParams)].pName := LBuffer;
            Result.fParams[High(Result.fParams)].pType := LType;

            SetLength(LBuffer, Length(AFuncHeader) - LCount);
            LCount := 0;

            LStatus := fhsWaitBeforeParamType;
          end;

          ')' : // end parameters
          begin
            SetLength(LBuffer, LCount);
            SetLength(Result.fParams, Succ(Length(Result.fParams)));
            Result.fParams[High(Result.fParams)].pName := LBuffer;
            Result.fParams[High(Result.fParams)].pType := LType;

            SetLength(LBuffer, Length(AFuncHeader) - LCount);
            LCount := 0;

            LStatus := fhsWaitAfterParam;
          end;
        else
          raise TFuncHeaderInfoException.Create(BuildParseError(AFuncHeader, LNext, LBuffer, LCount, CErrorUnknownNameCharacter));
        end;
      end;

      fhsWaitAfterParamName :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            // do nothing
          end;

          ',' : // type
          begin
            LStatus := fhsWaitBeforeParamType;
          end;

          ')' : // end parameters
          begin
            LStatus := fhsWaitAfterParam;
          end;
        else
          raise TFuncHeaderInfoException.Create(BuildParseError(AFuncHeader, LNext, LBuffer, LCount, CErrorUnknownCharacter));
        end;
      end;

      fhsWaitAfterParam :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            // do nothing
          end;

          '{' :
          begin
            LBracketCount := 1;

            Inc(LCount);
            LBuffer[LCount] := LNext;

            LStatus := fhsSource;
          end;

          ';' : // end
          begin
            LStatus := fhsWaitForError;
          end;
        else
          raise TFuncHeaderInfoException.Create(BuildParseError(AFuncHeader, LNext, LBuffer, LCount, CErrorUnknownCharacter));
        end;
      end;

      fhsSource :
      begin
        case LNext of
          '{' : // open bracket
          begin
            Inc(LBracketCount);

            Inc(LCount);
            LBuffer[LCount] := LNext;
          end;

          '}' : // close bracket
          begin
            Dec(LBracketCount);

            Inc(LCount);
            LBuffer[LCount] := LNext;

            if (LBracketCount <= 0) then
            begin
              SetLength(LBuffer, LCount);
              Result.FStatement := FetchNextStatement(LBuffer);

              SetLength(LBuffer, Length(AFuncHeader) - LCount);
              LCount := 0;

              LStatus := fhsWaitForEnd;
            end;
          end;
        else
          Inc(LCount);
          LBuffer[LCount] := LNext;
        end;
      end;

      fhsWaitForEnd :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            // do nothing
          end;

          ';' : // end
          begin
            LStatus := fhsWaitForError;
          end;
        else
          raise TFuncHeaderInfoException.Create(BuildParseError(AFuncHeader, LNext, LBuffer, LCount, CErrorUnknownCharacter));
        end;
      end;

      fhsWaitForError :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            // do nothing
          end;
        else
          raise TFuncHeaderInfoException.Create(BuildParseError(AFuncHeader, LNext, LBuffer, LCount, CErrorUnknownCharacter));
        end;
      end;
    end;
  end;

  if ((LStatus <> fhsWaitForError) and (LStatus <> fhsWaitForEnd)) then
    raise TFuncHeaderInfoException.Create(BuildParseError(AFuncHeader, LNext, LBuffer, LCount, CErrorNoEnding));
end;

end.
