unit StmtsU;

(*
//  C-to-Quadcode Compiler
//  ----------------------
//  Practical Work:
//      Compilers Summer Semester 2008
//      (Lecture by Mr. Schwinn)
//
//  Course:
//      TAI-06-ABC
//
//  Developers:
//     Kevin Niehage   (Matrikelnummer: 280888)
//     Michaela Hansel (Matrikelnummer: 221664)
*)

// IMPLEMENTATION STATUS: DONE

interface

uses
  ConstsU,
  ErrorsU,
  ExpU,
  TypesU,
  VarsU;

procedure EmptyStatementInfo(var AStatementInfo : TStatementInfo);
procedure FreeStatementInfo(var AStatementInfo : TStatementInfo);

function FetchNextStatement(var ASource : String) : TStatementInfo;

implementation

procedure EmptyStatementInfo(var AStatementInfo : TStatementInfo);
begin
  AStatementInfo.sElseStatements    := nil;
  AStatementInfo.sExpression.eEntry := nil;
  AStatementInfo.sType              := sitNone;
  SetLength(AStatementInfo.sStatements, 0);
  SetLength(AStatementInfo.sVariables, 0);
end;

procedure FreeStatementInfo(var AStatementInfo : TStatementInfo);
  procedure FreeStatementEntry(var AStatementEntry : PStatementInfo);
  var
    LIndex : LongInt;
  begin
    if (AStatementEntry <> nil) then
    begin
      FreeStatementEntry(AStatementEntry^.sElseStatements);
      FreeExpInfo(AStatementEntry^.sExpression);
      SetLength(AStatementEntry^.sVariables, 0);

      for LIndex := 0 to Pred(Length(AStatementEntry^.sStatements)) do
        FreeStatementEntry(AStatementEntry^.sStatements[LIndex]);

      Dispose(AStatementEntry);
      AStatementEntry := nil;
    end;
  end;
var
  LEntry : PStatementInfo;
begin
  New(LEntry);
  LEntry^ := AStatementInfo;
  FreeStatementEntry(LEntry);

  EmptyStatementInfo(AStatementInfo);
end;

function FetchNextStatement(var ASource : String) : TStatementInfo;
  function ReadMatchingBrackets(var ASource : String) : String;
  var
    LCount   : LongInt;
    LIndex   : LongInt;
    LNext    : Char;
    LStarted : Boolean;
  begin
    Result := '';

    LCount   := 0;
//    LNext    := #0;
    LStarted := false;
    for LIndex := 1 to Length(ASource) do
    begin
      LNext := ASource[LIndex];

      if LStarted then
      begin
        case LNext of
          '(' : Inc(LCount);
          ')' : Dec(LCount);
        end;

        if (LCount = 0) then
        begin
          Result := Copy(ASource, 1, LIndex);
          Delete(ASource, 1, LIndex);

          Break;
        end;
      end
      else
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            // do nothing
          end;

          '(' :
          begin
            Inc(LCount);

            LStarted := true;
          end;
        else
          raise TStatementInfoException.Create(CErrorUnknownCharacter + LNext);
        end;
      end;
    end;

    if not(LStarted and (LCount = 0)) then
      raise TStatementInfoException.Create(CErrorNoEnding);
  end;

  function ReadToSemicolon(var ASource : String) : String;
  var
    LIndex : LongInt;
  begin
    Result := '';

    LIndex := Pos(';', ASource);
    if (LIndex > 0) then
    begin
      Result := Copy(ASource, 1, Pred(LIndex));
      Delete(ASource, 1, LIndex);
    end
    else
      raise TStatementInfoException.Create(CErrorNoEnding);
  end;

var
  LBuffer : String;
  LCount  : LongInt;
  LIndex  : LongInt;
  LNext   : Char;
  LStatus : TStatementStatus;
  LTemp   : TStatementInfo;
begin
  Result.sElseStatements    := nil;
  Result.sExpression.eEntry := nil;
  Result.sType              := sitNone;
  SetLength(Result.sStatements, 0);
  SetLength(Result.sVariables, 0);

  SetLength(LBuffer, Length(ASource));
  LCount  := 0;

  LNext   := #0;
  LStatus := ssWaitBeforeFirst;

  LIndex := 1;
  while (LIndex <= Length(ASource)) do
  begin
    LNext := ASource[LIndex];

    case LStatus of
      ssWaitBeforeFirst :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            // do nothing
          end;

          ';' : // empty expression
          begin
            Result.sType := sitExpression;
            Delete(ASource, 1, LIndex);

            LStatus := ssDone;
          end;

          '{' : // block open
          begin
            Result.sType := sitBlock;
            Delete(ASource, 1, LIndex);

            repeat
              LTemp := FetchNextStatement(ASource);
              if not(LTemp.sType in [sitNone, sitBlock, sitBlockEnd]) then
              begin
                SetLength(Result.sStatements, Succ(Length(Result.sStatements)));

                New(Result.sStatements[High(Result.sStatements)]);
                Result.sStatements[High(Result.sStatements)]^ := LTemp;
              end
            until ((LTemp.sType = sitBlockEnd) or
                   (LTemp.sType = sitNone));

            if (LTemp.sType <> sitNone) then
              LStatus := ssDone
            else
              raise TStatementInfoException.Create(CErrorNoEnding);
          end;

          '}' : // block end
          begin
            Result.sType := sitBlockEnd;
            Delete(ASource, 1, LIndex);

            LStatus := ssDone;
          end;
        else
          Inc(LCount);
          LBuffer[LCount] := LNext;

          LStatus := ssFirst;
        end;
      end;

      ssFirst :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            SetLength(LBuffer, LCount);

            if ((LBuffer = CCBool) or
                (LBuffer = CCConst) or
                (LBuffer = CCFloat) or
                (LBuffer = CCInt)) then // variable definition
            begin
              Result.sType      := sitVariable;
              Result.sVariables := ParseVarDef(ReadToSemicolon(ASource) + ';'); // add deleted semicolon

              LStatus := ssDone;
            end
            else
            begin
              if (LBuffer = CCDo) then // do while
              begin
                Delete(ASource, 1, LIndex);

                Result.sType := sitDo;

                LTemp := FetchNextStatement(ASource);
                if (LTemp.sType <> sitNone) then
                begin
                  SetLength(Result.sStatements, 1);

                  New(Result.sStatements[High(Result.sStatements)]);
                  Result.sStatements[High(Result.sStatements)]^ := LTemp;

                  LCount := 0;
                  LIndex := 0;
                  SetLength(LBuffer, Length(ASource));

                  LStatus := ssDoBefore;
                end
                else
                  raise TStatementInfoException.Create(CErrorNoEnding);
              end
              else
              begin
                if (LBuffer = CCIf) then // if
                begin
                  Delete(ASource, 1, LIndex);

                  Result.sType       := sitIf;
                  Result.sExpression := ParseExpression(ReadMatchingBrackets(ASource));

                  LTemp := FetchNextStatement(ASource);
                  if (LTemp.sType <> sitNone) then
                  begin
                    SetLength(Result.sStatements, 1);

                    New(Result.sStatements[High(Result.sStatements)]);
                    Result.sStatements[High(Result.sStatements)]^ := LTemp;

                    LCount := 0;
                    LIndex := 0;
                    SetLength(LBuffer, Length(ASource));

                    LStatus := ssIfBefore;
                  end
                  else
                    raise TStatementInfoException.Create(CErrorNoEnding);
                end
                else
                begin
                  if (LBuffer = CCReturn) then // return
                  begin
                    Delete(ASource, 1, LIndex);

                    Result.sType       := sitReturn;
                    Result.sExpression := ParseExpression(ReadToSemicolon(ASource));

                    LStatus := ssDone;
                  end
                  else
                  begin
                    if (LBuffer = CCWhile) then // while
                    begin
                      Delete(ASource, 1, LIndex);

                      Result.sType       := sitWhile;
                      Result.sExpression := ParseExpression(ReadMatchingBrackets(ASource));

                      LTemp := FetchNextStatement(ASource);
                      if (LTemp.sType <> sitNone) then
                      begin
                        SetLength(Result.sStatements, 1);

                        New(Result.sStatements[High(Result.sStatements)]);
                        Result.sStatements[High(Result.sStatements)]^ := LTemp;

                        LStatus := ssDone;
                      end
                      else
                        raise TStatementInfoException.Create(CErrorNoEnding);
                    end
                    else
                    begin // expression
                      Result.sExpression := ParseExpression(ReadToSemicolon(ASource));
                      Result.sType       := sitExpression;

                      LStatus := ssDone;
                    end;
                  end;
                end;
              end;
            end;
          end;

          ';' : // "do;while(...);", "return;" or expression
          begin
            SetLength(LBuffer, LCount);

            if (LBuffer = CCDo) then // do while
            begin
              Delete(ASource, 1, Pred(LIndex));

              Result.sType := sitDo;

              LTemp := FetchNextStatement(ASource);
              if (LTemp.sType <> sitNone) then
              begin
                SetLength(Result.sStatements, 1);

                New(Result.sStatements[High(Result.sStatements)]);
                Result.sStatements[High(Result.sStatements)]^ := LTemp;

                LCount := 0;
                LIndex := 0;
                SetLength(LBuffer, Length(ASource));

                LStatus := ssDoBefore;
              end
              else
                raise TStatementInfoException.Create(CErrorNoEnding);
            end
            else
            begin
              if (LBuffer = CCReturn) then // return
              begin
                Delete(ASource, 1, Pred(LIndex));

                Result.sType       := sitReturn;
                Result.sExpression := ParseExpression(ReadToSemicolon(ASource));

                LStatus := ssDone;
              end
              else
              begin // expression
                Result.sExpression := ParseExpression(ReadToSemicolon(ASource));
                Result.sType       := sitExpression;

                LStatus := ssDone;
              end;
            end;
          end;

          '(' : // if(...), while(...), return(...) or expression
          begin
            SetLength(LBuffer, LCount);

            if (LBuffer = CCIf) then // if
            begin
              Delete(ASource, 1, Pred(LIndex));

              Result.sType       := sitIf;
              Result.sExpression := ParseExpression(ReadMatchingBrackets(ASource));

              LTemp := FetchNextStatement(ASource);
              if (LTemp.sType <> sitNone) then
              begin
                SetLength(Result.sStatements, 1);

                New(Result.sStatements[High(Result.sStatements)]);
                Result.sStatements[High(Result.sStatements)]^ := LTemp;

                LCount := 0;
                LIndex := 0;
                SetLength(LBuffer, Length(ASource));

                LStatus := ssIfBefore;
              end
              else
                raise TStatementInfoException.Create(CErrorNoEnding);
            end
            else
            begin
              if (LBuffer = CCReturn) then // return
              begin
                Delete(ASource, 1, Pred(LIndex));

                Result.sType       := sitReturn;
                Result.sExpression := ParseExpression(ReadToSemicolon(ASource));

                LStatus := ssDone;
              end
              else
              begin
                if (LBuffer = CCWhile) then // while
                begin
                  Delete(ASource, 1, Pred(LIndex));

                  Result.sType       := sitWhile;
                  Result.sExpression := ParseExpression(ReadMatchingBrackets(ASource));

                  LTemp := FetchNextStatement(ASource);
                  if (LTemp.sType <> sitNone) then
                  begin
                    SetLength(Result.sStatements, 1);

                    New(Result.sStatements[High(Result.sStatements)]);
                    Result.sStatements[High(Result.sStatements)]^ := LTemp;

                    LStatus := ssDone;
                  end
                  else
                    raise TStatementInfoException.Create(CErrorNoEnding);
                end
                else
                begin // expression
                  Result.sExpression := ParseExpression(ReadToSemicolon(ASource));
                  Result.sType       := sitExpression;

                  LStatus := ssDone;
                end;
              end;
            end;
          end;

          '{' :
          begin
            SetLength(LBuffer, LCount);

            if (LBuffer = CCDo) then // do while
            begin
              Delete(ASource, 1, Pred(LIndex));

              Result.sType := sitDo;

              LTemp := FetchNextStatement(ASource);
              if (LTemp.sType <> sitNone) then
              begin
                SetLength(Result.sStatements, 1);

                New(Result.sStatements[High(Result.sStatements)]);
                Result.sStatements[High(Result.sStatements)]^ := LTemp;

                LCount := 0;
                LIndex := 0;
                SetLength(LBuffer, Length(ASource));

                LStatus := ssDoBefore;
              end
              else
                raise TStatementInfoException.Create(CErrorNoEnding);
            end
            else
              raise TStatementInfoException.Create(BuildParseError(ASource, LNext, LBuffer, LCount, CErrorUnknownOperand));
          end;
        else
          Inc(LCount);
          LBuffer[LCount] := LNext;
        end;
      end;

      ssDoBefore :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            // do nothing
          end;
        else
          Inc(LCount);
          LBuffer[LCount] := LNext;

          LStatus := ssDo;
        end;
      end;

      ssDo :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            SetLength(LBuffer, LCount);

            if (LBuffer = CCWhile) then
            begin
              Delete(ASource, 1, LIndex);

              Result.sExpression := ParseExpression(ReadMatchingBrackets(ASource));

              LCount := 0;
              LIndex := 0;
              SetLength(LBuffer, Length(ASource));

              LStatus := ssDoAfter;
            end
            else
              raise TStatementInfoException.Create(BuildParseError(ASource, LNext, LBuffer, LCount, CErrorUnknownOperand));
          end;

          '(' : // expression
          begin
            SetLength(LBuffer, LCount);

            if (LBuffer = CCWhile) then
            begin
              Delete(ASource, 1, Pred(LIndex));

              Result.sExpression := ParseExpression(ReadMatchingBrackets(ASource));

              LCount := 0;
              LIndex := 0;
              SetLength(LBuffer, Length(ASource));

              LStatus := ssDoAfter;
            end
            else
              raise TStatementInfoException.Create(BuildParseError(ASource, LNext, LBuffer, LCount, CErrorUnknownOperand));
          end;
        else
          Inc(LCount);
          LBuffer[LCount] := LNext;
        end;
      end;

      ssDoAfter :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            // do nothing
          end;

          ';' : // end
          begin
            LStatus := ssDone;
          end;
        else
          raise TStatementInfoException.Create(BuildParseError(ASource, LNext, LBuffer, LCount, CErrorUnknownCharacter));
        end;
      end;

      ssIfBefore :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            // do nothing
          end;
        else
          Inc(LCount);
          LBuffer[LCount] := LNext;

          LStatus := ssIf;
        end;
      end;

      ssIf :
      begin
        case LNext of
          #9, #10, #13, #32, '{', ';' : // seperator
          begin
            SetLength(LBuffer, LCount);

            if (LBuffer = CCElse) then
            begin
              if (LNext in ['{', ';']) then
                Dec(LIndex);
              Delete(ASource, 1, LIndex);

              LTemp := FetchNextStatement(ASource);
              if (LTemp.sType <> sitNone) then
              begin
                New(Result.sElseStatements);
                Result.sElseStatements^ := LTemp;
              end
              else
                raise TStatementInfoException.Create(CErrorNoEnding);
            end;

            LStatus := ssDone;
          end;
        else
          Inc(LCount);
          LBuffer[LCount] := LNext;
        end;
      end;

      ssDone :
      begin
        Break;
      end;
    end;

    Inc(LIndex);
  end;

  if (LStatus <> ssDone) then
    raise TVarInfoException.Create(BuildParseError(ASource, LNext, LBuffer, LCount, CErrorNoEnding));
end;

end.
