unit SymbolsU;

(*
//  C-to-Quadcode Compiler
//  ----------------------
//  Practical Work:
//      Compilers Summer Semester 2008
//      (Lecture by Mr. Schwinn)
//
//  Course:
//      TAI-06-ABC
//
//  Developers:
//     Kevin Niehage   (Matrikelnummer: 280888)
//     Michaela Hansel (Matrikelnummer: 221664)
*)

// IMPLEMENTATION STATUS: DONE

interface

uses
  Classes,
  SysUtils,
  ConstsU,
  TypesU;

type
  TSymbolEntry = class;
  TSymbolTable = class;

  TSymbolEntry = class(TObject)
  private
  protected
    FNr        : LongInt;
    FName      : String;
    FTyp       : TSymbolTableType;
    FIntTyp    : TSymbolTableType;
    FOffset    : LongInt;
    FLine      : LongInt;
    FIndex1    : LongInt;
    FIndex2    : LongInt;
    FVater     : LongInt;
    FParameter : LongInt;
  public
    constructor Create(const ANr : LongInt; const AName : String; ATyp : TSymbolTableType;
                       const AIntTyp : TSymbolTableType; const AOffset : LongInt;
                       const ALine : LongInt; const AIndex1 : LongInt;
                       const AIndex2 : LongInt; const AVater : LongInt;
                       const AParameter : LongInt);

    property Nr        : LongInt          read FNr;
    property Name      : String           read FName;
    property Typ       : TSymbolTableType read FTyp;
    property IntTyp    : TSymbolTableType read FIntTyp;
    property Offset    : LongInt          read FOffset    write FOffset;
    property Line      : LongInt          read FLine      write FLine;
    property Index1    : LongInt          read FIndex1;
    property Index2    : LongInt          read FIndex2;
    property Vater     : LongInt          read FVater;
    property Parameter : LongInt          read FParameter;
  published
  end;

  TSymbolTable = class(TList)
  private
  protected
    function GetTyp(const AValue : TSymbolTableType; const AIntValue : TSymbolTableType) : String;
    function GetIntTyp(const AIntValue : TSymbolTableType; const AValue : TSymbolTableType) : String;
    function GetOffset : LongInt;

    procedure SetOffset(const AValue : LongInt);
  public
    constructor Create(const AProgram : String);

    property Offset : LongInt read GetOffset;

    function AddEntry(const AName : String; ATyp : TSymbolTableType;
                      const AIntTyp : TSymbolTableType; const ALine : LongInt;
                      const AVater : LongInt; const AParameter : LongInt) : LongInt;
    function EntryExists(const AName : String; const AVater : LongInt) : LongInt;
    function GetEntry(const ANr : LongInt) : TSymbolEntry;
    function ParamTypeList(const AName : String; const AVater : LongInt) : TSymbolTableTypeArray;

    function PrintSymbolTable : String;
  published
  end;

implementation

{ TSymbolEntry }

constructor TSymbolEntry.Create(const ANr : LongInt; const AName : String; ATyp : TSymbolTableType;
                                const AIntTyp : TSymbolTableType; const AOffset : LongInt;
                                const ALine : LongInt; const AIndex1 : LongInt;
                                const AIndex2 : LongInt; const AVater : LongInt;
                                const AParameter : LongInt);
begin
  inherited Create;

  FNr        := ANr;
  FName      := AName;
  FTyp       := ATyp;
  FIntTyp    := AIntTyp;
  FOffset    := AOffset;
  FLine      := ALine;
  FIndex1    := AIndex1;
  FIndex2    := AIndex2;
  FVater     := AVater;
  FParameter := AParameter;
end;

{ TSymbolList }

function TSymbolTable.AddEntry(const AName : String; ATyp : TSymbolTableType;
                               const AIntTyp : TSymbolTableType; const ALine : LongInt;
                               const AVater : LongInt; const AParameter : LongInt) : LongInt;
  function GetOffset(const AOffset : LongInt; const ATyp : TSymbolTableType) : LongInt;
  begin
    Result := AOffset;

    case ATyp of
      sttBool : Result := Result + 1;
      sttInt  : Result := Result + 4;
      sttReal : Result := Result + 8;
    end;
  end;
var
  LOffset : LongInt;
begin
//  Result := -1;

  if ((AName = CCBool) or
      (AName = CCConst) or
      (AName = CCElse) or
      (AName = CCFloat) or
      (AName = CCInt) or
      (AName = CCReturn) or
      (AName = CCVoid)) then
    raise TSymbolTableException.Create(CErrorReservedWordAsIdentifier + AName);

  if ((ATyp <> sttNone) and
      (AIntTyp <> sttNone)) then
    raise TSymbolTableException.Create(CErrorUnknownSymbolType + 'Typ=' + GetTyp(ATyp, AIntTyp) + ', IntTyp=' + GetIntTyp(AIntTyp, ATyp));

  if (EntryExists(AName, AVater) < 0) then
  begin
    LOffset := 0;
    if ((ATyp <> sttNone) or (AIntTyp <> sttNone)) then
      LOffset := Offset;

    Result := Add(TSymbolEntry.Create(Count, AName, ATyp, AIntTyp, LOffset, ALine,
                                      0, 0, AVater, AParameter));

    if (ATyp <> sttNone) then
      SetOffset(GetOffset(Offset, ATyp))
    else
    begin
      if (AIntTyp <> sttNone) then
        SetOffset(GetOffset(Offset, AIntTyp));
    end;
  end
  else
    raise TSymbolTableException.Create(CErrorIdentifierAlreadyDefined + AName);
end;

function TSymbolTable.EntryExists(const AName : String; const AVater : LongInt) : LongInt;
var
  LIndex : LongInt;
begin
  Result := - 1;

  for LIndex := 0 to Pred(Count) do
  begin
    if ((GetEntry(LIndex).Vater = AVater) and
        (GetEntry(LIndex).Name = AName)) then
    begin
      if ((GetEntry(LIndex).FTyp = sttNone) or
          (GetEntry(LIndex).FIntTyp = sttNone)) then
        Result := LIndex;
    end;

    if (Result >= 0) then
      Break;
  end;
end;

constructor TSymbolTable.Create(const AProgram : String);
begin
  inherited Create;

  Clear;
  Add(TSymbolEntry.Create(0, AProgram, sttBool, sttBool, 0, 0, 0, 0, 0, 0));
end;

function TSymbolTable.GetIntTyp(const AIntValue : TSymbolTableType; const AValue : TSymbolTableType) : String;
begin
  Result := '';

  case AIntValue of
    sttNone : Result := CQuadNone;

    sttBool :
    begin
      if (AValue = sttNone) then
        Result := CQuadBool
      else
        Result := CQuadNone;
    end;

    sttInt :
    begin
      if (AValue = sttNone) then
        Result := CQuadInt
      else
        Result := CQuadNone;
    end;

    sttReal :
    begin
      if (AValue = sttNone) then
        Result := CQuadReal
      else
        Result := CQuadNone;
    end;
  end;
end;

function TSymbolTable.GetTyp(const AValue : TSymbolTableType; const AIntValue : TSymbolTableType) : String;
begin
  Result := '';

  case AValue of
    sttNone :
    begin
      if (AIntValue = sttNone) then
        Result := CQuadProc
      else
        Result := CQuadFunc;
    end;

    sttBool :
    begin
      if (AIntValue = sttNone) then
        Result := CQuadBool
      else
        Result := CQuadPrg;
    end;

    sttInt :
    begin
      if (AIntValue = sttNone) then
        Result := CQuadInt
      else
        Result := CQuadPrg;
    end;

    sttReal :
    begin
      if (AIntValue = sttNone) then
        Result := CQuadReal
      else
        Result := CQuadPrg;
    end;
  end;
end;

function TSymbolTable.GetEntry(const ANr : LongInt) : TSymbolEntry;
begin
  Result := nil;
  if ((ANr >= 0) and (ANr < Count)) then
    Result := Items[ANr];
end;

function TSymbolTable.GetOffset : LongInt;
begin
  Result := 0;
  if (Count > 0) then
    Result := GetEntry(0).Offset;
end;

function TSymbolTable.PrintSymbolTable : String;
var
  LIndex : LongInt;
begin
  Result := '';

  for LIndex := 0 to Pred(Count) do
    Result := Result +
              IntToStr(GetEntry(LIndex).Nr) + ':' + #9 +
              GetEntry(LIndex).Name + #9 +
              GetTyp(GetEntry(LIndex).Typ, GetEntry(LIndex).IntTyp) + #9 +
              GetIntTyp(GetEntry(LIndex).IntTyp, GetEntry(LIndex).Typ) + #9 +
              IntToStr(GetEntry(LIndex).Offset) + #9 +
              IntToStr(GetEntry(LIndex).Line) + #9 +
              IntToStr(GetEntry(LIndex).Index1) + #9 +
              IntToStr(GetEntry(LIndex).Index2) + #9 +
              IntToStr(GetEntry(LIndex).Vater) + #9 +
              IntToStr(GetEntry(LIndex).Parameter) + sLineBreak;
end;

procedure TSymbolTable.SetOffset(const AValue : Integer);
begin
  if (Count > 0) then
    GetEntry(0).Offset := AValue;
end;

function TSymbolTable.ParamTypeList(const AName: String; const AVater: Integer): TSymbolTableTypeArray;
var
  LEntryIndex : LongInt;
  LIndex      : LongInt;
begin
  SetLength(Result, 0);

  LEntryIndex := EntryExists(AName, AVater);
  if (LEntryIndex > 0) then
  begin
    if (GetEntry(LEntryIndex).Typ = sttNone) then
    begin
      SetLength(Result, GetEntry(LEntryIndex).Parameter);

      for LIndex := 1 to Pred(Count) do
      begin
        if (GetEntry(LIndex).Vater = LEntryIndex) and
           (GetEntry(LIndex).Parameter > 0) then
          Result[Pred(GetEntry(LIndex).Parameter)] := GetEntry(LIndex).Typ;
      end;
    end;
  end;
end;

end.
