unit TypesU;

(*
//  C-to-Quadcode Compiler
//  ----------------------
//  Practical Work:
//      Compilers Summer Semester 2008
//      (Lecture by Mr. Schwinn)
//
//  Course:
//      TAI-06-ABC
//
//  Developers:
//     Kevin Niehage   (Matrikelnummer: 280888)
//     Michaela Hansel (Matrikelnummer: 221664)
*)

// IMPLEMENTATION STATUS: DONE

interface

uses
  SysUtils;

type
  TCharArray   = array of Char;
  TStringArray = array of String;

// BEGIN: Compiler
  TCompilerException = class(Exception);
// END: Compiler

// BEGIN: ExpInfo
  TExpValueOption = (evoNone, evoPreInc, evoPreDec, evoPostInc, evoPostDec);
  TExpValueType   = (evtNone, evtFloat, evtFunctionCall, evtInteger, evtHex,
                     evtOct, evtVariable);

  TExpOperation = (eoNone,
                   eoPostDecVal, eoPostIncVal,
                   eoNotLog, eoNotVal, eoPreDecVal, eoPreIncVal, eoUnaryMinusVal, eoUnaryPlusVal,
                   eoDivVal, eoModVal, eoMulVal,
                   eoMinusVal, eoPlusVal,
                   eoShlVal, eoShrVal,
                   eoGreaterEqualLog, eoGreaterLog, eoLessEqualLog, eoLessLog, 
                   eoEqualLog, eoNotEqualLog,
                   eoAndVal,
                   eoXorVal,
                   eoOrVal,
                   eoAndLog,
                   eoOrLog,
                   {eoAndAssign, }eoAssign{, eoDivAssign, eoMinusAssign,
                   eoModAssign, eoMulAssign, eoOrAssign, eoPlusAssign,
                   eoShlAssign, eoShrAssign, eoXorAssign});

  PExpEntry = ^TExpEntry;
  
  TExpInfo = record
    eEntry : PExpEntry;
  end;
  TExpInfoArray = array of TExpInfo;

  TExpValue = record
    vOption : TExpValueOption;
    vParams : TExpInfoArray;
    vType   : TExpValueType;
    vValue  : String;
  end;
  TExpValueArray = array of TExpValue;

  TExpEntry = record
    eBrackets  : LongInt;
    eLeft      : PExpEntry;
    eOperation : TExpOperation;
    eRight     : PExpEntry;
    eValues    : TExpValueArray;
  end;

  TExpInfoException = class(Exception);

  TExpStatus = (esWaitBefore, esInteger, esFloat, esHexOrOct, esHex, esHexBeforeNum,
                esOct, esVariable, esFunctionCall, esOperator, esWriteOperator,
                esExponentBeforeSign, esExponentBeforeNum, esExponent);
// END: ExpInfo

// BEGIN: FileInfo
  TFileInfoType = (fitForward, fitFunction, fitVariable);

  TFileInfo = record
    fSource : String;
    fType   : TFileInfoType;
  end;
  TFileInfoArray = array of TFileInfo;

  TFileInfoException = class(Exception);

  TFileStatus = (fsWaitForNext, fsWaitForEnd, fsWaitForVariableEnd,
                 fsWaitForForwardEnd, fsWaitForFunctionEnd, fsWriteEntry);
// END: FileInfo

// BEGIN : FuncCallInfo
  TFuncCallInfo = record
    fName   : String;
    fParams : TStringArray;
  end;

  TFuncCallInfoException = class(Exception);

  TFuncCallStatus = (fcsWaitBeforeName, fcsName, fcsWaitAfterName, fcsParam,
                     fcsWaitForError);
// END: FuncCallInfo

// BEGIN: InterValue
    TInterValueConstruct = (ivcNone, ivcValue, ivcFuncCall, ivcVariable);
    TInterValueType      = (ivtNone, ivtInt, ivtFloat);
    TInterValue = record
      iConstruct : TInterValueConstruct;
      iType      : TInterValueType;
      iValue     : String;
    end;
    TInterValueArray = array of TInterValue;
// END: InterValue

// BEGIN: Quotes
  TQuoteException = class(Exception);

  TQuoteStatus = (qsWaitForNext, qsQuoteFirst, qsLineQuote, qsBlockQuote,
                  qsBlockQuoteAfterFirst);
// END  : Quotes

// BEGIN: Symbol
  TSymbolType = (stFunction, stVariable);
  TSymbol = record
    sName : String;
    sType : TSymbolType;
  end;
  TSymbolArray = array of TSymbol;
// END : Symbol

// BEGIN: SymbolTable
  TSymbolTableType = (sttNone, sttBool, sttInt, sttReal);
  TSymbolTableTypeArray = array of TSymbolTableType;

  TSymbolTableException = class(Exception);
// END: SymbolTable

// BEGIN: Switches
  TSwitchStatus = (ssWaitForSwitch, ssSwitch);
// END  : Switches

// BEGIN: VarInfo
  TVarInfo = record
    vConst : Boolean;
    vName  : String;
    vType  : TSymbolTableType;
    vValue : TExpInfo;
  end;
  TVarInfoArray = array of TVarInfo;

  TVarInfoException = class(Exception);

  TVarStatus = (vsWaitBeforeType, vsType, vsWaitBeforeName, vsName,
                vsWaitAfterName, vsWaitBeforeValue, vsValue, vsWaitForError);
// END: VarInfo

// BEGIN: StatementInfo
  TStatementInfoType = (sitNone, sitBlock, sitBlockEnd, sitVariable, sitExpression,
                        sitWhile, sitDo, sitIf, sitReturn);

  PStatementInfo = ^TStatementInfo;
  TPStatementInfoArray = array of PStatementInfo;

  TStatementInfo = record
    sElseStatements : PStatementInfo;
    sExpression     : TExpInfo;
    sStatements     : TPStatementInfoArray;
    sType           : TStatementInfoType;
    sVariables      : TVarInfoArray;
  end;
  TStatementInfoArray = array of TStatementInfo;

  TStatementInfoException = class(Exception);

  TStatementStatus = (ssWaitBeforeFirst, ssFirst, ssDoBefore, ssDo, ssDoAfter,
                      ssIfBefore, ssIf, ssDone);
// END: StatementInfo

// BEGIN: FuncHeaderInfo
  TParamInfo = record
    pName : String;
    pType : String;
  end;
  TParamInfoArray = array of TParamInfo;

  TFuncHeaderInfo = record
    fName      : String;
    fType      : TSymbolTableType;
    fParams    : TParamInfoArray;
    fStatement : TStatementInfo;
  end;
  TFuncHeaderInfoArray = array of TFuncHeaderInfo;

  TFuncHeaderInfoException = class(Exception);

  TFuncHeaderStatus = (fhsWaitBeforeType, fhsType, fhsWaitBeforeName, fhsName,
                       fhsWaitAfterName, fhsWaitBeforeParamType, fhsParamType,
                       fhsWaitAfterParamType, fhsWaitBeforeParamName,
                       fhsParamName, fhsWaitAfterParamName, fhsWaitAfterParam,
                       fhsSource, fhsWaitForEnd, fhsWaitForError);
// END: FuncHeaderInfo

implementation

end.
