unit VarsU;

(*
//  C-to-Quadcode Compiler
//  ----------------------
//  Practical Work:
//      Compilers Summer Semester 2008
//      (Lecture by Mr. Schwinn)
//
//  Course:
//      TAI-06-ABC
//
//  Developers:
//     Kevin Niehage   (Matrikelnummer: 280888)
//     Michaela Hansel (Matrikelnummer: 221664)
*)

// IMPLEMENTATION STATUS: DONE

interface

uses
  SysUtils,
  ConstsU,
  ErrorsU,
  ExpU,
  TypesU;

function GetVariableType(AValue : String) : TSymbolTableType;

function ParseVarDef(const AVarDef : String) : TVarInfoArray;

implementation

function GetVariableType(AValue : String) : TSymbolTableType;
begin
//  Result := sttNone;

  AValue := Trim(AValue);
  if (AValue = CCBool) then
    Result := sttInt // C bool is replaced by Quad int
  else
  begin
    if (AValue = CCInt) then
      Result := sttInt
    else
    begin
      if (AValue = CCFloat) then
        Result := sttReal
      else
      begin
        if (AValue = CCVoid) then
          Result := sttNone
        else
          raise TVarInfoException.Create(CErrorUnknownVariableType + AValue);
      end;
    end;
  end;
end;

function ParseVarDef(const AVarDef : String) : TVarInfoArray;
var
  LBuffer : String;
  LConst  : Boolean;
  LCount  : LongInt;
  LExp    : TExpInfo;
  LIndex  : LongInt;
  LNext   : Char;
  LStatus : TVarStatus;
  LType   : String;
  LValue  : String;
begin
  SetLength(Result, 0);

  SetLength(LBuffer, Length(AVarDef));
  LCount  := 0;

  LConst      := false;
  LExp.eEntry := nil;
  LNext       := #0;
  LStatus     := vsWaitBeforeType;
  LType       := '';
  LValue      := '';
  for LIndex := 1 to Length(AVarDef) do
  begin
    LNext := AVarDef[LIndex];

    case LStatus of
      vsWaitBeforeType :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            // do nothing
          end;

          'A'..'Z', 'a'..'z' : // type
          begin
            Inc(LCount);
            LBuffer[LCount] := LNext;

            LStatus := vsType;
          end;
        else
          raise TVarInfoException.Create(BuildParseError(AVarDef, LNext, LBuffer, LCount, CErrorUnknownTypeCharacter));
        end;
      end;

      vsType :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            SetLength(LBuffer, LCount);

            if (LBuffer = CCConst) then
            begin
              LConst := true;

              LStatus := vsWaitBeforeType;
            end
            else
            begin
              LType := LBuffer;

              LStatus := vsWaitBeforeName;
            end;

            SetLength(LBuffer, Length(AVarDef) - LCount);
            LCount := 0;
          end;

          'A'..'Z', 'a'..'z' : // type
          begin
            Inc(LCount);
            LBuffer[LCount] := LNext;
          end;
        else
          raise TVarInfoException.Create(BuildParseError(AVarDef, LNext, LBuffer, LCount, CErrorUnknownTypeCharacter));
        end;
      end;

      vsWaitBeforeName :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            // do nothing
          end;

          'A'..'Z', 'a'..'z', '_' : // name
          begin
            Inc(LCount);
            LBuffer[LCount] := LNext;

            LStatus := vsName;
          end;
        else
          raise TVarInfoException.Create(BuildParseError(AVarDef, LNext, LBuffer, LCount, CErrorUnknownNameCharacter));
        end;
      end;

      vsName :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            SetLength(LBuffer, LCount);

            if (LBuffer = CCConst) then
            begin
              LConst := true;

              LStatus := vsWaitBeforeName
            end
            else
            begin
              SetLength(Result, Succ(Length(Result)));
              Result[High(Result)].vConst        := LConst;
              Result[High(Result)].vName         := LBuffer;
              Result[High(Result)].vType         := GetVariableType(LType);
              Result[High(Result)].vValue.eEntry := nil;

              LStatus := vsWaitAfterName;
            end;

            SetLength(LBuffer, Length(AVarDef) - LCount);
            LCount := 0;
          end;

          'A'..'Z', 'a'..'z', '_', '0'..'9' : // name
          begin
            Inc(LCount);
            LBuffer[LCount] := LNext;
          end;

          ',' : // name seperator
          begin
            SetLength(LBuffer, LCount);
            SetLength(Result, Succ(Length(Result)));
            Result[High(Result)].vConst        := LConst;
            Result[High(Result)].vName         := LBuffer;
            Result[High(Result)].vType         := GetVariableType(LType);
            Result[High(Result)].vValue.eEntry := nil;

            SetLength(LBuffer, Length(AVarDef) - LCount);
            LCount := 0;

            LStatus := vsWaitBeforeName;
          end;

          '=' : // value seperator
          begin
            SetLength(LBuffer, LCount);
            SetLength(Result, Succ(Length(Result)));
            Result[High(Result)].vConst        := LConst;
            Result[High(Result)].vName         := LBuffer;
            Result[High(Result)].vType         := GetVariableType(LType);
            Result[High(Result)].vValue.eEntry := nil;

            SetLength(LBuffer, Length(AVarDef) - LCount);
            LCount := 0;

            LStatus := vsValue;
          end;

          ';' : // end
          begin
            SetLength(LBuffer, LCount);
            SetLength(Result, Succ(Length(Result)));
            Result[High(Result)].vConst        := LConst;
            Result[High(Result)].vName         := LBuffer;
            Result[High(Result)].vType         := GetVariableType(LType);
            Result[High(Result)].vValue.eEntry := nil;

            SetLength(LBuffer, Length(AVarDef) - LCount);
            LCount := 0;

            LStatus := vsWaitForError;
          end;
        else
          raise TVarInfoException.Create(BuildParseError(AVarDef, LNext, LBuffer, LCount, CErrorUnknownNameCharacter));
        end;
      end;

      vsWaitAfterName :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            // do nothing
          end;

          ',' : // name seperator
          begin
            LStatus := vsWaitBeforeName;
          end;

          '=' : // value
          begin
            LStatus := vsValue;
          end;

          ';' : // end
          begin
            LStatus := vsWaitForError;
          end;
        else
          raise TVarInfoException.Create(BuildParseError(AVarDef, LNext, LBuffer, LCount, CErrorUnknownNameCharacter));
        end;
      end;

      vsValue :
      begin
        case LNext of
          ';' : // end
          begin
            SetLength(LBuffer, LCount);
            LValue := LBuffer;

            SetLength(LBuffer, Length(AVarDef) - LCount);
            LCount := 0;

            LStatus := vsWaitForError;
          end;
        else
          Inc(LCount);
          LBuffer[LCount] := LNext;
        end;
      end;

      vsWaitForError :
      begin
        case LNext of
          #9, #10, #13, #32 : // seperator
          begin
            // do nothing
          end;
        else
          raise TVarInfoException.Create(BuildParseError(AVarDef, LNext, LBuffer, LCount, CErrorUnknownCharacter));
        end;
      end;
    end;
  end;

  if (LStatus <> vsWaitForError) then
    raise TVarInfoException.Create(BuildParseError(AVarDef, LNext, LBuffer, LCount, CErrorNoEnding));

  if (Length(LValue) <= 0) then
    LValue := '0';

  LExp := ParseExpression(LValue);
  for LIndex := 0 to Pred(Length(Result)) do
    Result[LIndex].vValue := LExp;
end;

end.
